from __future__ import annotations

import abc
import logging
import os
import sys
import urllib.parse
import urllib.request
import warnings
from typing import Any, Iterator, Literal, Optional, Union

import sqlalchemy as sql
from tqdm import TqdmWarning, tqdm

from pixeltable import catalog, exceptions as excs, exprs
from pixeltable.env import Env
from pixeltable.exec import ExecNode
from pixeltable.metadata import schema
from pixeltable.utils.media_store import MediaStore
from pixeltable.utils.sql import log_explain, log_stmt

_logger = logging.getLogger('pixeltable')


class StoreBase:
    """Base class for stored tables

    Each row has the following system columns:
    - rowid columns: one or more columns that identify a user-visible row across all versions
    - v_min: version at which the row was created
    - v_max: version at which the row was deleted (or MAX_VERSION if it's still live)
    """

    tbl_version: catalog.TableVersionHandle
    sa_md: sql.MetaData
    sa_tbl: Optional[sql.Table]
    _pk_cols: list[sql.Column]
    v_min_col: sql.Column
    v_max_col: sql.Column
    base: Optional[StoreBase]

    __INSERT_BATCH_SIZE = 1000

    def __init__(self, tbl_version: catalog.TableVersion):
        self.tbl_version = catalog.TableVersionHandle(
            tbl_version.id, tbl_version.effective_version, tbl_version=tbl_version
        )
        self.sa_md = sql.MetaData()
        self.sa_tbl = None
        # We need to declare a `base` variable here, even though it's only defined for instances of `StoreView`,
        # since it's referenced by various methods of `StoreBase`
        self.base = tbl_version.base.get().store_tbl if tbl_version.base is not None else None
        self.create_sa_tbl()

    def pk_columns(self) -> list[sql.Column]:
        return self._pk_cols

    def rowid_columns(self) -> list[sql.Column]:
        return self._pk_cols[:-1]

    @abc.abstractmethod
    def _create_rowid_columns(self) -> list[sql.Column]:
        """Create and return rowid columns"""

    def _create_system_columns(self) -> list[sql.Column]:
        """Create and return system columns"""
        rowid_cols = self._create_rowid_columns()
        self.v_min_col = sql.Column('v_min', sql.BigInteger, nullable=False)
        self.v_max_col = sql.Column(
            'v_max', sql.BigInteger, nullable=False, server_default=str(schema.Table.MAX_VERSION)
        )
        self._pk_cols = [*rowid_cols, self.v_min_col]
        return [*rowid_cols, self.v_min_col, self.v_max_col]

    def create_sa_tbl(self) -> None:
        """Create self.sa_tbl from self.tbl_version."""
        system_cols = self._create_system_columns()
        all_cols = system_cols.copy()
        for col in [c for c in self.tbl_version.get().cols if c.is_stored]:
            # re-create sql.Column for each column, regardless of whether it already has sa_col set: it was bound
            # to the last sql.Table version we created and cannot be reused
            col.create_sa_cols()
            all_cols.append(col.sa_col)
            if col.records_errors:
                all_cols.append(col.sa_errormsg_col)
                all_cols.append(col.sa_errortype_col)

        if self.sa_tbl is not None:
            # if we're called in response to a schema change, we need to remove the old table first
            self.sa_md.remove(self.sa_tbl)

        idxs: list[sql.Index] = []
        # index for all system columns:
        # - base x view joins can be executed as merge joins
        # - speeds up ORDER BY rowid DESC
        # - allows filtering for a particular table version in index scan
        idx_name = f'sys_cols_idx_{self.tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, *system_cols))

        # v_min/v_max indices: speeds up base table scans needed to propagate a base table insert or delete
        idx_name = f'vmin_idx_{self.tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, self.v_min_col, postgresql_using=Env.get().dbms.version_index_type))
        idx_name = f'vmax_idx_{self.tbl_version.id.hex}'
        idxs.append(sql.Index(idx_name, self.v_max_col, postgresql_using=Env.get().dbms.version_index_type))

        self.sa_tbl = sql.Table(self._storage_name(), self.sa_md, *all_cols, *idxs)

    @abc.abstractmethod
    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        """Return predicate for rowid joins to all bases"""

    @abc.abstractmethod
    def _storage_name(self) -> str:
        """Return the name of the data store table"""

    def _move_tmp_media_file(self, file_url: Optional[str], col: catalog.Column, v_min: int) -> str:
        """Move tmp media file with given url to Env.media_dir and return new url, or given url if not a tmp_dir file"""
        pxt_tmp_dir = str(Env.get().tmp_dir)
        if file_url is None:
            return None
        parsed = urllib.parse.urlparse(file_url)
        # We should never be passed a local file path here. The "len > 1" ensures that Windows
        # file paths aren't mistaken for URLs with a single-character scheme.
        assert len(parsed.scheme) > 1
        if parsed.scheme != 'file':
            # remote url
            return file_url
        file_path = urllib.parse.unquote(urllib.request.url2pathname(parsed.path))
        if not file_path.startswith(pxt_tmp_dir):
            # not a tmp file
            return file_url
        _, ext = os.path.splitext(file_path)
        new_path = str(MediaStore.prepare_media_path(self.tbl_version.id, col.id, v_min, ext=ext))
        os.rename(file_path, new_path)
        new_file_url = urllib.parse.urljoin('file:', urllib.request.pathname2url(new_path))
        return new_file_url

    def _move_tmp_media_files(
        self, table_rows: list[dict[str, Any]], media_cols: list[catalog.Column], v_min: int
    ) -> None:
        """Move tmp media files that we generated to a permanent location"""
        for c in media_cols:
            for table_row in table_rows:
                file_url = table_row[c.store_name()]
                table_row[c.store_name()] = self._move_tmp_media_file(file_url, c, v_min)

    def _create_table_row(
        self, input_row: exprs.DataRow, row_builder: exprs.RowBuilder, exc_col_ids: set[int], pk: tuple[int, ...]
    ) -> tuple[dict[str, Any], int]:
        """Return Tuple[complete table row, # of exceptions] for insert()
        Creates a row that includes the PK columns, with the values from input_row.pk.
        Returns:
            Tuple[complete table row, # of exceptions]
        """
        table_row, num_excs = row_builder.create_table_row(input_row, exc_col_ids)
        assert len(pk) == len(self._pk_cols)
        for pk_col, pk_val in zip(self._pk_cols, pk):
            table_row[pk_col.name] = pk_val
        return table_row, num_excs

    def count(self) -> int:
        """Return the number of rows visible in self.tbl_version"""
        stmt = (
            sql.select(sql.func.count('*'))
            .select_from(self.sa_tbl)
            .where(self.v_min_col <= self.tbl_version.get().version)
            .where(self.v_max_col > self.tbl_version.get().version)
        )
        conn = Env.get().conn
        result = conn.execute(stmt).scalar_one()
        assert isinstance(result, int)
        return result

    def create(self) -> None:
        conn = Env.get().conn
        self.sa_md.create_all(bind=conn)

    def drop(self) -> None:
        """Drop store table"""
        conn = Env.get().conn
        self.sa_md.drop_all(bind=conn)

    def add_column(self, col: catalog.Column) -> None:
        """Add column(s) to the store-resident table based on a catalog column

        Note that a computed catalog column will require two extra columns (for the computed value and for the error
        message).
        """
        assert col.is_stored
        conn = Env.get().conn
        col_type_str = col.get_sa_col_type().compile(dialect=conn.dialect)
        stmt = sql.text(f'ALTER TABLE {self._storage_name()} ADD COLUMN {col.store_name()} {col_type_str} NULL')
        log_stmt(_logger, stmt)
        conn.execute(stmt)
        added_storage_cols = [col.store_name()]
        if col.records_errors:
            # we also need to create the errormsg and errortype storage cols
            stmt = sql.text(
                f'ALTER TABLE {self._storage_name()} ADD COLUMN {col.errormsg_store_name()} VARCHAR DEFAULT NULL'
            )
            conn.execute(stmt)
            stmt = sql.text(
                f'ALTER TABLE {self._storage_name()} ADD COLUMN {col.errortype_store_name()} VARCHAR DEFAULT NULL'
            )
            conn.execute(stmt)
            added_storage_cols.extend([col.errormsg_store_name(), col.errortype_store_name()])
        self.create_sa_tbl()
        _logger.info(f'Added columns {added_storage_cols} to storage table {self._storage_name()}')

    def drop_column(self, col: catalog.Column) -> None:
        """Execute Alter Table Drop Column statement"""
        conn = Env.get().conn
        stmt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.store_name()}'
        conn.execute(sql.text(stmt))
        if col.records_errors:
            stmt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.errormsg_store_name()}'
            conn.execute(sql.text(stmt))
            stmt = f'ALTER TABLE {self._storage_name()} DROP COLUMN {col.errortype_store_name()}'
            conn.execute(sql.text(stmt))

    def load_column(
        self, col: catalog.Column, exec_plan: ExecNode, value_expr_slot_idx: int, on_error: Literal['abort', 'ignore']
    ) -> int:
        """Update store column of a computed column with values produced by an execution plan

        Returns:
            number of rows with exceptions
        Raises:
            sql.exc.DBAPIError if there was a SQL error during execution
            excs.Error if on_error='abort' and there was an exception during row evaluation
        """
        assert col.tbl.id == self.tbl_version.id
        num_excs = 0
        num_rows = 0

        # create temp table to store output of exec_plan, with the same primary key as the store table
        tmp_name = f'temp_{self._storage_name()}'
        tmp_pk_cols = [sql.Column(col.name, col.type, primary_key=True) for col in self.pk_columns()]
        tmp_cols = tmp_pk_cols.copy()
        tmp_val_col = sql.Column(col.sa_col.name, col.sa_col.type)
        tmp_cols.append(tmp_val_col)
        # add error columns if the store column records errors
        if col.records_errors:
            tmp_errortype_col = sql.Column(col.sa_errortype_col.name, col.sa_errortype_col.type)
            tmp_cols.append(tmp_errortype_col)
            tmp_errormsg_col = sql.Column(col.sa_errormsg_col.name, col.sa_errormsg_col.type)
            tmp_cols.append(tmp_errormsg_col)
        tmp_tbl = sql.Table(tmp_name, self.sa_md, *tmp_cols, prefixes=['TEMPORARY'])
        conn = Env.get().conn
        tmp_tbl.create(bind=conn)

        try:
            # insert rows from exec_plan into temp table
            # TODO: unify the table row construction logic with RowBuilder.create_table_row()
            for row_batch in exec_plan:
                num_rows += len(row_batch)
                tbl_rows: list[dict[str, Any]] = []
                for result_row in row_batch:
                    tbl_row: dict[str, Any] = {}
                    for pk_col, pk_val in zip(self.pk_columns(), result_row.pk):
                        tbl_row[pk_col.name] = pk_val

                    if col.is_computed:
                        if result_row.has_exc(value_expr_slot_idx):
                            num_excs += 1
                            value_exc = result_row.get_exc(value_expr_slot_idx)
                            if on_error == 'abort':
                                raise excs.Error(
                                    f'Error while evaluating computed column `{col.name}`:\n{value_exc}'
                                ) from value_exc
                            # we store a NULL value and record the exception/exc type
                            error_type = type(value_exc).__name__
                            error_msg = str(value_exc)
                            tbl_row[col.sa_col.name] = None
                            tbl_row[col.sa_errortype_col.name] = error_type
                            tbl_row[col.sa_errormsg_col.name] = error_msg
                        else:
                            if col.col_type.is_image_type() and result_row.file_urls[value_expr_slot_idx] is None:
                                # we have yet to store this image
                                filepath = str(MediaStore.prepare_media_path(col.tbl.id, col.id, col.tbl.get().version))
                                result_row.flush_img(value_expr_slot_idx, filepath)
                            val = result_row.get_stored_val(value_expr_slot_idx, col.sa_col.type)
                            if col.col_type.is_media_type():
                                val = self._move_tmp_media_file(val, col, result_row.pk[-1])
                            tbl_row[col.sa_col.name] = val
                            if col.records_errors:
                                tbl_row[col.sa_errortype_col.name] = None
                                tbl_row[col.sa_errormsg_col.name] = None

                    tbl_rows.append(tbl_row)
                conn.execute(sql.insert(tmp_tbl), tbl_rows)

            # update store table with values from temp table
            update_stmt = sql.update(self.sa_tbl)
            for pk_col, tmp_pk_col in zip(self.pk_columns(), tmp_pk_cols):
                update_stmt = update_stmt.where(pk_col == tmp_pk_col)
            update_stmt = update_stmt.values({col.sa_col: tmp_val_col})
            if col.records_errors:
                update_stmt = update_stmt.values(
                    {col.sa_errortype_col: tmp_errortype_col, col.sa_errormsg_col: tmp_errormsg_col}
                )
            log_explain(_logger, update_stmt, conn)
            conn.execute(update_stmt)

        finally:
            tmp_tbl.drop(bind=conn)
            self.sa_md.remove(tmp_tbl)
        return num_excs

    def insert_rows(
        self,
        exec_plan: ExecNode,
        v_min: Optional[int] = None,
        show_progress: bool = True,
        rowids: Optional[Iterator[int]] = None,
        abort_on_exc: bool = False,
    ) -> tuple[int, int, set[int]]:
        """Insert rows into the store table and update the catalog table's md
        Returns:
            number of inserted rows, number of exceptions, set of column ids that have exceptions
        """
        assert v_min is not None
        # TODO: total?
        num_excs = 0
        num_rows = 0
        cols_with_excs: set[int] = set()
        progress_bar: Optional[tqdm] = None  # create this only after we started executing
        row_builder = exec_plan.row_builder
        media_cols = [info.col for info in row_builder.table_columns if info.col.col_type.is_media_type()]
        conn = Env.get().conn

        try:
            exec_plan.open()
            for row_batch in exec_plan:
                num_rows += len(row_batch)
                for batch_start_idx in range(0, len(row_batch), self.__INSERT_BATCH_SIZE):
                    # compute batch of rows and convert them into table rows
                    table_rows: list[dict[str, Any]] = []
                    batch_stop_idx = min(batch_start_idx + self.__INSERT_BATCH_SIZE, len(row_batch))
                    for row_idx in range(batch_start_idx, batch_stop_idx):
                        row = row_batch[row_idx]
                        # if abort_on_exc == True, we need to check for media validation exceptions
                        if abort_on_exc and row.has_exc():
                            exc = row.get_first_exc()
                            raise exc

                        rowid = (next(rowids),) if rowids is not None else row.pk[:-1]
                        pk = (*rowid, v_min)
                        table_row, num_row_exc = self._create_table_row(row, row_builder, cols_with_excs, pk=pk)
                        num_excs += num_row_exc
                        table_rows.append(table_row)

                        if show_progress:
                            if progress_bar is None:
                                warnings.simplefilter('ignore', category=TqdmWarning)
                                progress_bar = tqdm(
                                    desc=f'Inserting rows into `{self.tbl_version.get().name}`',
                                    unit=' rows',
                                    ncols=100,
                                    file=sys.stdout,
                                )
                            progress_bar.update(1)

                    # insert batch of rows
                    self._move_tmp_media_files(table_rows, media_cols, v_min)
                    conn.execute(sql.insert(self.sa_tbl), table_rows)
            if progress_bar is not None:
                progress_bar.close()
            return num_rows, num_excs, cols_with_excs
        finally:
            exec_plan.close()

    def _versions_clause(self, versions: list[Optional[int]], match_on_vmin: bool) -> sql.ColumnElement[bool]:
        """Return filter for base versions"""
        v = versions[0]
        if v is None:
            # we're looking at live rows
            clause = sql.and_(
                self.v_min_col <= self.tbl_version.get().version, self.v_max_col == schema.Table.MAX_VERSION
            )
        else:
            # we're looking at a specific version
            clause = self.v_min_col == v if match_on_vmin else self.v_max_col == v
        if len(versions) == 1:
            return clause
        return sql.and_(clause, self.base._versions_clause(versions[1:], match_on_vmin))

    def delete_rows(
        self,
        current_version: int,
        base_versions: list[Optional[int]],
        match_on_vmin: bool,
        where_clause: Optional[sql.ColumnElement[bool]],
    ) -> int:
        """Mark rows as deleted that are live and were created prior to current_version.
        Also: populate the undo columns
        Args:
            base_versions: if non-None, join only to base rows that were created at that version,
                otherwise join to rows that are live in the base's current version (which is distinct from the
                current_version parameter)
            match_on_vmin: if True, match exact versions on v_min; if False, match on v_max
            where_clause: if not None, also apply where_clause
        Returns:
            number of deleted rows
        """
        where_clause = sql.true() if where_clause is None else where_clause
        where_clause = sql.and_(
            self.v_min_col < current_version, self.v_max_col == schema.Table.MAX_VERSION, where_clause
        )
        rowid_join_clause = self._rowid_join_predicate()
        base_versions_clause = (
            sql.true() if len(base_versions) == 0 else self.base._versions_clause(base_versions, match_on_vmin)
        )
        set_clause: dict[sql.Column, Union[int, sql.Column]] = {self.v_max_col: current_version}
        for index_info in self.tbl_version.get().idxs_by_name.values():
            # copy value column to undo column
            set_clause[index_info.undo_col.sa_col] = index_info.val_col.sa_col
            # set value column to NULL
            set_clause[index_info.val_col.sa_col] = None
        stmt = (
            sql.update(self.sa_tbl)
            .values(set_clause)
            .where(where_clause)
            .where(rowid_join_clause)
            .where(base_versions_clause)
        )
        conn = Env.get().conn
        log_explain(_logger, stmt, conn)
        status = conn.execute(stmt)
        return status.rowcount


class StoreTable(StoreBase):
    def __init__(self, tbl_version: catalog.TableVersion):
        assert not tbl_version.is_view
        super().__init__(tbl_version)

    def _create_rowid_columns(self) -> list[sql.Column]:
        self.rowid_col = sql.Column('rowid', sql.BigInteger, nullable=False)
        return [self.rowid_col]

    def _storage_name(self) -> str:
        return f'tbl_{self.tbl_version.id.hex}'

    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        return sql.true()


class StoreView(StoreBase):
    def __init__(self, catalog_view: catalog.TableVersion):
        assert catalog_view.is_view
        super().__init__(catalog_view)

    def _create_rowid_columns(self) -> list[sql.Column]:
        # a view row corresponds directly to a single base row, which means it needs to duplicate its rowid columns
        self.rowid_cols = [sql.Column(c.name, c.type) for c in self.base.rowid_columns()]
        return self.rowid_cols

    def _storage_name(self) -> str:
        return f'view_{self.tbl_version.id.hex}'

    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        return sql.and_(
            self.base._rowid_join_predicate(),
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns(), self.base.rowid_columns())],
        )


class StoreComponentView(StoreView):
    """A view that stores components of its base, as produced by a ComponentIterator

    PK: now also includes pos, the position returned by the ComponentIterator for the base row identified by base_rowid
    """

    rowid_cols: list[sql.Column]
    pos_col: sql.Column
    pos_col_idx: int

    def __init__(self, catalog_view: catalog.TableVersion):
        super().__init__(catalog_view)

    def _create_rowid_columns(self) -> list[sql.Column]:
        # each base row is expanded into n view rows
        self.rowid_cols = [sql.Column(c.name, c.type) for c in self.base.rowid_columns()]
        # name of pos column: avoid collisions with bases' pos columns
        self.pos_col = sql.Column(f'pos_{len(self.rowid_cols) - 1}', sql.BigInteger, nullable=False)
        self.pos_col_idx = len(self.rowid_cols)
        self.rowid_cols.append(self.pos_col)
        return self.rowid_cols

    def create_sa_tbl(self) -> None:
        super().create_sa_tbl()
        # we need to fix up the 'pos' column in TableVersion
        self.tbl_version.get().cols_by_name['pos'].sa_col = self.pos_col

    def _rowid_join_predicate(self) -> sql.ColumnElement[bool]:
        return sql.and_(
            self.base._rowid_join_predicate(),
            *[c1 == c2 for c1, c2 in zip(self.rowid_columns()[:-1], self.base.rowid_columns())],
        )



from __future__ import annotations

import inspect
import logging
from typing import TYPE_CHECKING, Any, List, Literal, Optional
from uuid import UUID

import pixeltable.exceptions as excs
import pixeltable.metadata.schema as md_schema
import pixeltable.type_system as ts
from pixeltable import exprs, func
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator

from .column import Column
from .globals import _POS_COLUMN_NAME, MediaValidation, UpdateStatus
from .table import Table
from .table_version import TableVersion
from .table_version_handle import TableVersionHandle
from .table_version_path import TableVersionPath

if TYPE_CHECKING:
    import pixeltable as pxt

_logger = logging.getLogger('pixeltable')


class View(Table):
    """A `Table` that presents a virtual view of another table (or view).

    A view is typically backed by a store table, which records the view's columns and is joined back to the bases
    at query execution time.
    The exception is a snapshot view without a predicate and without additional columns: in that case, the view
    is simply a reference to a specific set of base versions.
    """

    def __init__(self, id: UUID, dir_id: UUID, name: str, tbl_version_path: TableVersionPath, snapshot_only: bool):
        super().__init__(id, dir_id, name, tbl_version_path)
        self._snapshot_only = snapshot_only

    @classmethod
    def _display_name(cls) -> str:
        return 'view'

    @classmethod
    def select_list_to_additional_columns(cls, select_list: list[tuple[exprs.Expr, Optional[str]]]) -> dict[str, dict]:
        """Returns a list of columns in the same format as the additional_columns parameter of View.create.
        The source is the list of expressions from a select() statement on a DataFrame.
        If the column is a ColumnRef, to a base table column, it is marked to not be stored.sy
        """
        from pixeltable.dataframe import DataFrame

        r: dict[str, dict] = {}
        exps, names = DataFrame._normalize_select_list([], select_list)
        for expr, name in zip(exps, names):
            stored = not isinstance(expr, exprs.ColumnRef)
            r[name] = {'value': expr, 'stored': stored}
        return r

    @classmethod
    def _create(
        cls,
        dir_id: UUID,
        name: str,
        base: TableVersionPath,
        select_list: Optional[list[tuple[exprs.Expr, Optional[str]]]],
        additional_columns: dict[str, Any],
        predicate: Optional['pxt.exprs.Expr'],
        is_snapshot: bool,
        num_retained_versions: int,
        comment: str,
        media_validation: MediaValidation,
        iterator_cls: Optional[type[ComponentIterator]],
        iterator_args: Optional[dict],
    ) -> View:
        # Convert select_list to more additional_columns if present
        include_base_columns: bool = select_list is None
        select_list_columns: List[Column] = []
        if not include_base_columns:
            r = cls.select_list_to_additional_columns(select_list)
            select_list_columns = cls._create_columns(r)

        columns_from_additional_columns = cls._create_columns(additional_columns)
        columns = select_list_columns + columns_from_additional_columns
        cls._verify_schema(columns)

        # verify that filter can be evaluated in the context of the base
        if predicate is not None:
            if not predicate.is_bound_by([base]):
                raise excs.Error(f'Filter cannot be computed in the context of the base {base.tbl_name()}')
            # create a copy that we can modify and store
            predicate = predicate.copy()

        # same for value exprs
        for col in columns:
            if not col.is_computed:
                continue
            # make sure that the value can be computed in the context of the base
            if col.value_expr is not None and not col.value_expr.is_bound_by([base]):
                raise excs.Error(
                    f'Column {col.name}: value expression cannot be computed in the context of the '
                    f'base {base.tbl_name()}'
                )

        if iterator_cls is not None:
            assert iterator_args is not None

            # validate iterator_args
            py_signature = inspect.signature(iterator_cls.__init__)

            # make sure iterator_args can be used to instantiate iterator_cls
            bound_args: dict[str, Any]
            try:
                bound_args = py_signature.bind(None, **iterator_args).arguments  # None: arg for self
            except TypeError as exc:
                raise excs.Error(f'Invalid iterator arguments: {exc}') from exc
            # we ignore 'self'
            first_param_name = next(iter(py_signature.parameters))  # can't guarantee it's actually 'self'
            del bound_args[first_param_name]

            # construct Signature and type-check bound_args
            params = [
                func.Parameter(param_name, param_type, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
                for param_name, param_type in iterator_cls.input_schema().items()
            ]
            sig = func.Signature(ts.InvalidType(), params)

            expr_args = {k: exprs.Expr.from_object(v) for k, v in bound_args.items()}
            sig.validate_args(expr_args, context=f'in iterator {iterator_cls.__name__!r}')
            literal_args = {k: v.val if isinstance(v, exprs.Literal) else v for k, v in expr_args.items()}

            # prepend pos and output_schema columns to cols:
            # a component view exposes the pos column of its rowid;
            # we create that column here, so it gets assigned a column id;
            # stored=False: it is not stored separately (it's already stored as part of the rowid)
            iterator_cols = [Column(_POS_COLUMN_NAME, ts.IntType(), stored=False)]
            output_dict, unstored_cols = iterator_cls.output_schema(**literal_args)
            iterator_cols.extend(
                [
                    Column(col_name, col_type, stored=col_name not in unstored_cols)
                    for col_name, col_type in output_dict.items()
                ]
            )

            iterator_col_names = {col.name for col in iterator_cols}
            for col in columns:
                if col.name in iterator_col_names:
                    raise excs.Error(
                        f'Duplicate name: column {col.name} is already present in the iterator output schema'
                    )
            columns = iterator_cols + columns

        session = Env.get().session
        from pixeltable.exprs import InlineDict

        iterator_args_expr: exprs.Expr = InlineDict(iterator_args) if iterator_args is not None else None
        iterator_class_fqn = f'{iterator_cls.__module__}.{iterator_cls.__name__}' if iterator_cls is not None else None
        base_version_path = cls._get_snapshot_path(base) if is_snapshot else base

        # if this is a snapshot, we need to retarget all exprs to the snapshot tbl versions
        if is_snapshot:
            predicate = predicate.retarget(base_version_path) if predicate is not None else None
            iterator_args_expr = (
                iterator_args_expr.retarget(base_version_path) if iterator_args_expr is not None else None
            )
            for col in columns:
                if col.value_expr is not None:
                    col.set_value_expr(col.value_expr.retarget(base_version_path))

        view_md = md_schema.ViewMd(
            is_snapshot=is_snapshot,
            include_base_columns=include_base_columns,
            predicate=predicate.as_dict() if predicate is not None else None,
            base_versions=base_version_path.as_md(),
            iterator_class_fqn=iterator_class_fqn,
            iterator_args=iterator_args_expr.as_dict() if iterator_args_expr is not None else None,
        )

        id, tbl_version = TableVersion.create(
            dir_id,
            name,
            columns,
            num_retained_versions,
            comment,
            media_validation=media_validation,
            # base_path=base_version_path,
            view_md=view_md,
        )
        if tbl_version is None:
            # this is purely a snapshot: we use the base's tbl version path
            view = cls(id, dir_id, name, base_version_path, snapshot_only=True)
            _logger.info(f'created snapshot {name}')
        else:
            view = cls(
                id,
                dir_id,
                name,
                TableVersionPath(
                    TableVersionHandle(tbl_version.id, tbl_version.effective_version), base=base_version_path
                ),
                snapshot_only=False,
            )
            _logger.info(f'Created view `{name}`, id={tbl_version.id}')

            from pixeltable.plan import Planner

            plan, _ = Planner.create_view_load_plan(view._tbl_version_path)
            num_rows, num_excs, _ = tbl_version.store_tbl.insert_rows(plan, v_min=tbl_version.version)
            Env.get().console_logger.info(f'Created view `{name}` with {num_rows} rows, {num_excs} exceptions.')

        session.commit()
        return view

    @classmethod
    def _verify_column(cls, col: Column) -> None:
        # make sure that columns are nullable or have a default
        if not col.col_type.nullable and not col.is_computed:
            raise excs.Error(f'Column {col.name}: non-computed columns in views must be nullable')
        super()._verify_column(col)

    @classmethod
    def _get_snapshot_path(cls, tbl_version_path: TableVersionPath) -> TableVersionPath:
        """Returns snapshot of the given table version path.
        All TableVersions of that path will be snapshot versions. Creates new versions from mutable versions,
        if necessary.
        """
        if tbl_version_path.is_snapshot():
            return tbl_version_path
        tbl_version = tbl_version_path.tbl_version.get()
        if not tbl_version.is_snapshot:
            # create and register snapshot version
            tbl_version = tbl_version.create_snapshot_copy()
            assert tbl_version.is_snapshot

        return TableVersionPath(
            TableVersionHandle(tbl_version.id, tbl_version.effective_version),
            base=cls._get_snapshot_path(tbl_version_path.base) if tbl_version_path.base is not None else None,
        )

    def _drop(self) -> None:
        if self._snapshot_only:
            # there is not TableVersion to drop
            self._check_is_dropped()
            self.is_dropped = True
            TableVersion.delete_md(self._id)
        else:
            super()._drop()

    def get_metadata(self) -> dict[str, Any]:
        md = super().get_metadata()
        md['is_view'] = True
        md['is_snapshot'] = self._tbl_version_path.is_snapshot()
        return md

    if TYPE_CHECKING:
        import datasets  # type: ignore[import-untyped]

        from pixeltable.globals import RowData, TableDataSource

    def insert(
        self,
        source: Optional[TableDataSource] = None,
        /,
        *,
        source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
        schema_overrides: Optional[dict[str, ts.ColumnType]] = None,
        on_error: Literal['abort', 'ignore'] = 'abort',
        print_stats: bool = False,
        **kwargs: Any,
    ) -> UpdateStatus:
        raise excs.Error(f'{self._display_name()} {self._name!r}: cannot insert into view')

    def delete(self, where: Optional[exprs.Expr] = None) -> UpdateStatus:
        raise excs.Error(f'{self._display_name()} {self._name!r}: cannot delete from view')



import copy
import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterator, Literal, Optional, cast
from xml.etree import ElementTree as ET

import label_studio_sdk  # type: ignore[import-untyped]
import PIL.Image
from requests.exceptions import HTTPError

import pixeltable as pxt
from pixeltable import Column, Table, env, exceptions as excs
from pixeltable.config import Config
from pixeltable.exprs import ColumnRef, DataRow, Expr
from pixeltable.io.external_store import Project, SyncStatus
from pixeltable.utils import coco

# label_studio_sdk>=1 and label_studio_sdk<1 are not compatible, so we need to try
# the import two different ways to insure intercompatibility
try:
    # label_studio_sdk<1 compatibility
    import label_studio_sdk.project as ls_project  # type: ignore
except ImportError:
    # label_studio_sdk>=1 compatibility
    import label_studio_sdk._legacy.project as ls_project  # type: ignore

_logger = logging.getLogger('pixeltable')


@env.register_client('label_studio')
def _(api_key: str, url: str) -> label_studio_sdk.Client:
    return label_studio_sdk.Client(api_key=api_key, url=url)


def _label_studio_client() -> label_studio_sdk.Client:
    return env.Env.get().get_client('label_studio')


class LabelStudioProject(Project):
    """
    An [`ExternalStore`][pixeltable.io.ExternalStore] that represents a Label Studio project, providing functionality
    for synchronizing between a Pixeltable table and a Label Studio project.
    """

    def __init__(
        self,
        name: str,
        project_id: int,
        media_import_method: Literal['post', 'file', 'url'],
        col_mapping: dict[Column, str],
        stored_proxies: Optional[dict[Column, Column]] = None,
    ):
        """
        The constructor will NOT create a new Label Studio project; it is also used when loading
        metadata for existing projects.
        """
        self.project_id = project_id
        self.media_import_method = media_import_method
        self._project: Optional[ls_project.Project] = None
        super().__init__(name, col_mapping, stored_proxies)

    @property
    def project(self) -> ls_project.Project:
        """The `Project` object corresponding to this Label Studio project."""
        if self._project is None:
            try:
                self._project = _label_studio_client().get_project(self.project_id)
            except HTTPError as exc:
                raise excs.Error(
                    f'Could not locate Label Studio project: {self.project_id} '
                    '(cannot connect to server or project no longer exists)'
                ) from exc
        return self._project

    @property
    def project_params(self) -> dict[str, Any]:
        """The parameters of this Label Studio project."""
        return self.project.get_params()

    @property
    def project_title(self) -> str:
        """The title of this Label Studio project."""
        return self.project_params['title']

    @property
    def __project_config(self) -> '_LabelStudioConfig':
        return self.__parse_project_config(self.project_params['label_config'])

    def get_export_columns(self) -> dict[str, pxt.ColumnType]:
        """
        The data keys and preannotation fields specified in this Label Studio project.
        """
        return self.__project_config.export_columns

    def get_import_columns(self) -> dict[str, pxt.ColumnType]:
        """
        Always contains a single entry:

        ```
        {"annotations": pxt.JsonType(nullable=True)}
        ```
        """
        return {ANNOTATIONS_COLUMN: pxt.JsonType(nullable=True)}

    def sync(self, t: Table, export_data: bool, import_data: bool) -> SyncStatus:
        _logger.info(
            f'Syncing Label Studio project "{self.project_title}" with table `{t._name}`'
            f' (export: {export_data}, import: {import_data}).'
        )
        # Collect all existing tasks into a dict with entries `rowid: task`
        tasks = {tuple(task['meta']['rowid']): task for task in self.__fetch_all_tasks()}
        sync_status = SyncStatus.empty()
        if export_data:
            export_sync_status = self.__update_tasks(t, tasks)
            sync_status = sync_status.combine(export_sync_status)
        if import_data:
            import_sync_status = self.__update_table_from_tasks(t, tasks)
            sync_status = sync_status.combine(import_sync_status)
        return sync_status

    def __fetch_all_tasks(self) -> Iterator[dict[str, Any]]:
        """Retrieves all tasks and task metadata in this Label Studio project."""
        page = 1
        unknown_task_count = 0
        while True:
            result = self.project.get_paginated_tasks(page=page, page_size=_PAGE_SIZE)
            if result.get('end_pagination'):
                break
            for task in result['tasks']:
                rowid = task['meta'].get('rowid')
                if rowid is None:
                    unknown_task_count += 1
                else:
                    yield task
            page += 1
        if unknown_task_count > 0:
            _logger.warning(
                f'Skipped {unknown_task_count} unrecognized task(s) when syncing '
                f'Label Studio project {self.project_title!r}.'
            )

    def __update_tasks(self, t: Table, existing_tasks: dict[tuple, dict]) -> SyncStatus:
        """
        Updates all tasks in this Label Studio project based on the Pixeltable data:
        - Creates new tasks for rows that don't map to any existing task;
        - Updates existing tasks for rows whose data has changed;
        - Deletes any tasks whose rows no longer exist in the Pixeltable table.
        """
        config = self.__project_config

        # Columns in `t` that map to Label Studio data keys
        t_data_cols = [t_col for t_col, ext_col_name in self.col_mapping.items() if ext_col_name in config.data_keys]

        if len(t_data_cols) == 0:
            return SyncStatus.empty()

        # Columns in `t` that map to `rectanglelabels` preannotations
        t_rl_cols = [
            t_col for t_col, ext_col_name in self.col_mapping.items() if ext_col_name in config.rectangle_labels
        ]

        # Destinations for `rectanglelabels` preannotations
        rl_info = list(config.rectangle_labels.values())

        _logger.debug('`t_data_cols`: %s', t_data_cols)
        _logger.debug('`t_rl_cols`: %s', t_rl_cols)
        _logger.debug('`rl_info`: %s', rl_info)

        if self.media_import_method == 'post':
            # Send media to Label Studio by HTTP post.
            assert len(t_data_cols) == 1  # This was verified when the project was set up
            return self.__update_tasks_by_post(t, existing_tasks, t_data_cols[0], t_rl_cols, rl_info)
        elif self.media_import_method in ('file', 'url'):
            # Send media to Label Studio by file reference (local file or URL).
            return self.__update_tasks_by_files(t, existing_tasks, t_data_cols, t_rl_cols, rl_info)
        else:
            raise AssertionError()

    def __update_tasks_by_post(
        self,
        t: Table,
        existing_tasks: dict[tuple, dict],
        media_col: Column,
        t_rl_cols: list[Column],
        rl_info: list['_RectangleLabel'],
    ) -> SyncStatus:
        is_stored = media_col.is_stored
        # If it's a stored column, we can use `localpath`
        localpath_col_opt = [t[media_col.name].localpath] if is_stored else []
        # Select the media column, rectanglelabels columns, and localpath (if appropriate)
        rows = t.select(t[media_col.name], *[t[col.name] for col in t_rl_cols], *localpath_col_opt)
        tasks_created = 0
        row_ids_in_pxt: set[tuple] = set()

        for row in rows._exec():
            media_col_idx = rows._select_list_exprs[0].slot_idx
            rl_col_idxs = [expr.slot_idx for expr in rows._select_list_exprs[1 : 1 + len(t_rl_cols)]]
            row_ids_in_pxt.add(row.rowid)
            if row.rowid not in existing_tasks:
                # Upload the media file to Label Studio
                if is_stored:
                    # There is an existing localpath; use it!
                    localpath_col_idx = rows._select_list_exprs[-1].slot_idx
                    file = Path(row[localpath_col_idx])
                    task_id: int = self.project.import_tasks(file)[0]
                else:
                    # No localpath; create a temp file and upload it
                    assert isinstance(row[media_col_idx], PIL.Image.Image)
                    file = env.Env.get().create_tmp_path(extension='.png')
                    row[media_col_idx].save(file, format='png')
                    task_id = self.project.import_tasks(file)[0]
                    os.remove(file)

                # Update the task with `rowid` metadata
                self.project.update_task(task_id, meta={'rowid': row.rowid})

                # Convert coco annotations to predictions
                coco_annotations = [row[i] for i in rl_col_idxs]
                _logger.debug('`coco_annotations`: %s', coco_annotations)
                predictions = [
                    self.__coco_to_predictions(
                        coco_annotations[i], self.col_mapping[t_rl_cols[i]], rl_info[i], task_id=task_id
                    )
                    for i in range(len(coco_annotations))
                ]
                _logger.debug('`predictions`: {%s}', predictions)
                self.project.create_predictions(predictions)
                tasks_created += 1

        env.Env.get().console_logger.info(f'Created {tasks_created} new task(s) in {self}.')

        sync_status = SyncStatus(external_rows_created=tasks_created)

        deletion_sync_status = self.__delete_stale_tasks(existing_tasks, row_ids_in_pxt, tasks_created)

        return sync_status.combine(deletion_sync_status)

    def __update_tasks_by_files(
        self,
        t: Table,
        existing_tasks: dict[tuple, dict],
        t_data_cols: list[Column],
        t_rl_cols: list[Column],
        rl_info: list['_RectangleLabel'],
    ) -> SyncStatus:
        ext_data_cols = [self.col_mapping[col] for col in t_data_cols]
        expr_refs: dict[str, Expr] = {}  # kwargs for the select statement
        for col in t_data_cols:
            col_name = col.name
            if self.media_import_method == 'url':
                expr_refs[col_name] = t[col_name].fileurl
            else:
                assert self.media_import_method == 'file'
                if not col.col_type.is_media_type():
                    # Not a media column; query the data directly
                    expr_refs[col_name] = cast(ColumnRef, t[col_name])
                elif col in self.stored_proxies:
                    # Media column that has a stored proxy; use it. We have to give it a name,
                    # since it's an anonymous column
                    stored_proxy_col = self.stored_proxies[col]
                    expr_refs[f'{col_name}_proxy'] = ColumnRef(stored_proxy_col).localpath
                else:
                    # Media column without a stored proxy; this means it's a stored computed column,
                    # and we can just use the localpath
                    expr_refs[col_name] = t[col_name].localpath

        df = t.select(*[t[col.name] for col in t_rl_cols], **expr_refs)
        # The following buffers will hold `DataRow` indices that correspond to each of the selected
        # columns. `rl_col_idxs` holds the indices for the columns that map to RectangleLabels
        # preannotations; `data_col_idxs` holds the indices for the columns that map to data fields.
        # We have to wait until we begin iterating to populate them, so they're initially `None`.
        rl_col_idxs: Optional[list[int]] = None
        data_col_idxs: Optional[list[int]] = None

        row_ids_in_pxt: set[tuple] = set()
        tasks_created = 0
        tasks_updated = 0
        page: list[dict[str, Any]] = []  # buffer to hold tasks for paginated API calls

        # Function that turns a `DataRow` into a `dict` for creating or updating a task in the
        # Label Studio SDK.
        def create_task_info(row: DataRow) -> dict[str, Any]:
            data_vals = [row[idx] for idx in data_col_idxs]
            coco_annotations = [row[idx] for idx in rl_col_idxs]
            for i in range(len(t_data_cols)):
                if t_data_cols[i].col_type.is_media_type():
                    # Special handling for media columns
                    assert isinstance(data_vals[i], str)
                    if self.media_import_method == 'url':
                        data_vals[i] = self.__validate_fileurl(t_data_cols[i], data_vals[i])
                    else:
                        assert self.media_import_method == 'file'
                        data_vals[i] = self.__localpath_to_lspath(data_vals[i])
            predictions = [
                self.__coco_to_predictions(coco_annotations[i], self.col_mapping[t_rl_cols[i]], rl_info[i])
                for i in range(len(coco_annotations))
            ]
            return {
                'data': dict(zip(ext_data_cols, data_vals)),
                'meta': {'rowid': row.rowid},
                'predictions': predictions,
            }

        for row in df._exec():
            if rl_col_idxs is None:
                rl_col_idxs = [expr.slot_idx for expr in df._select_list_exprs[: len(t_rl_cols)]]
                data_col_idxs = [expr.slot_idx for expr in df._select_list_exprs[len(t_rl_cols) :]]
            row_ids_in_pxt.add(row.rowid)
            task_info = create_task_info(row)
            # TODO(aaron-siegel): Implement more efficient update logic (currently involves a full table scan)
            if row.rowid in existing_tasks:
                # A task for this row already exists; see if it needs an update.
                existing_task = existing_tasks[row.rowid]
                if (
                    task_info['data'] != existing_task['data']
                    or task_info['predictions'] != existing_task['predictions']
                ):
                    _logger.debug(f'Updating task for rowid {row.rowid}.')
                    self.project.update_task(existing_tasks[row.rowid]['id'], **task_info)
                    tasks_updated += 1
            else:
                # No task exists for this row; we need to create one.
                page.append(task_info)
                tasks_created += 1
                if len(page) == _PAGE_SIZE:
                    self.project.import_tasks(page)
                    page.clear()

        if len(page) > 0:
            self.project.import_tasks(page)

        env.Env.get().console_logger.info(
            f'Created {tasks_created} new task(s) and updated {tasks_updated} existing task(s) in {self}.'
        )

        sync_status = SyncStatus(external_rows_created=tasks_created, external_rows_updated=tasks_updated)

        deletion_sync_status = self.__delete_stale_tasks(existing_tasks, row_ids_in_pxt, tasks_created)

        return sync_status.combine(deletion_sync_status)

    @classmethod
    def __validate_fileurl(cls, col: Column, url: str) -> Optional[str]:
        # Check that the URL is one that will be visible to Label Studio. If it isn't, log an info message
        # to help users debug the issue.
        if not (url.startswith('http://') or url.startswith('https://')):
            _logger.info(
                f'URL found in media column `{col.name}` will not render correctly in Label Studio, since '
                f'it is not an HTTP URL: {url}'
            )
        return url

    @classmethod
    def __localpath_to_lspath(cls, localpath: str) -> str:
        # Transform the local path into Label Studio's bespoke path format.
        relpath = Path(localpath).relative_to(Config.get().home)
        return f'/data/local-files/?d={relpath}'

    def __delete_stale_tasks(
        self, existing_tasks: dict[tuple, dict], row_ids_in_pxt: set[tuple], tasks_created: int
    ) -> SyncStatus:
        deleted_rowids = set(existing_tasks.keys()) - row_ids_in_pxt
        # Sanity check the math
        assert len(deleted_rowids) == len(existing_tasks) + tasks_created - len(row_ids_in_pxt)
        tasks_to_delete = [existing_tasks[rowid]['id'] for rowid in deleted_rowids]

        if len(tasks_to_delete) > 0:
            self.project.delete_tasks(tasks_to_delete)
            env.Env.get().console_logger.info(
                f'Deleted {len(tasks_to_delete)} tasks(s) in {self} that are no longer present in Pixeltable.'
            )

        # Remove them from the `existing_tasks` dict so that future updates are applied correctly
        for rowid in deleted_rowids:
            del existing_tasks[rowid]

        return SyncStatus(external_rows_deleted=len(deleted_rowids))

    def __update_table_from_tasks(self, t: Table, tasks: dict[tuple, dict]) -> SyncStatus:
        if ANNOTATIONS_COLUMN not in self.col_mapping.values():
            return SyncStatus.empty()

        annotations = {
            # Replace [] by None to indicate no annotations. We do want to sync rows with no annotations,
            # in order to properly handle the scenario where existing annotations have been deleted in
            # Label Studio.
            tuple(task['meta']['rowid']): task[ANNOTATIONS_COLUMN] if len(task[ANNOTATIONS_COLUMN]) > 0 else None
            for task in tasks.values()
        }

        local_annotations_col = next(k for k, v in self.col_mapping.items() if v == ANNOTATIONS_COLUMN)

        # Prune the annotations down to just the ones that have actually changed.
        rows = t.select(t[local_annotations_col.name])
        for row in rows._exec():
            assert len(row.vals) == 1
            if row.rowid in annotations and annotations[row.rowid] == row[0]:
                del annotations[row.rowid]

        # Apply updates
        updates = [{'_rowid': rowid, local_annotations_col.name: ann} for rowid, ann in annotations.items()]
        if len(updates) > 0:
            _logger.info(
                f'Updating table {t._name!r}, column {local_annotations_col.name!r} '
                f'with {len(updates)} total annotations.'
            )
            # batch_update currently doesn't propagate from views to base tables. As a workaround, we call
            # batch_update on the actual ancestor table that holds the annotations column.
            # TODO(aaron-siegel): Simplify this once propagation is properly implemented in batch_update
            ancestor = t
            while local_annotations_col not in ancestor._tbl_version.get().cols:
                assert ancestor._base is not None
                ancestor = ancestor._base
            update_status = ancestor.batch_update(updates)
            env.Env.get().console_logger.info(f'Updated annotation(s) from {len(updates)} task(s) in {self}.')
            return SyncStatus(pxt_rows_updated=update_status.num_rows, num_excs=update_status.num_excs)
        else:
            return SyncStatus.empty()

    def as_dict(self) -> dict[str, Any]:
        return {
            'name': self.name,
            'project_id': self.project_id,
            'media_import_method': self.media_import_method,
            'col_mapping': [[self._column_as_dict(k), v] for k, v in self.col_mapping.items()],
            'stored_proxies': [
                [self._column_as_dict(k), self._column_as_dict(v)] for k, v in self.stored_proxies.items()
            ],
        }

    @classmethod
    def from_dict(cls, md: dict[str, Any]) -> 'LabelStudioProject':
        return LabelStudioProject(
            md['name'],
            md['project_id'],
            md['media_import_method'],
            {cls._column_from_dict(entry[0]): entry[1] for entry in md['col_mapping']},
            {cls._column_from_dict(entry[0]): cls._column_from_dict(entry[1]) for entry in md['stored_proxies']},
        )

    def __repr__(self) -> str:
        name = self.project.get_params()['title']
        return f'LabelStudioProject `{name}`'

    @classmethod
    def __parse_project_config(cls, xml_config: str) -> '_LabelStudioConfig':
        """
        Parses a Label Studio XML config, extracting the names and Pixeltable types of
        all input variables.
        """
        root: ET.Element = ET.fromstring(xml_config)
        if root.tag.lower() != 'view':
            raise excs.Error('Root of Label Studio config must be a `View`')
        config = _LabelStudioConfig(
            data_keys=cls.__parse_data_keys_config(root), rectangle_labels=cls.__parse_rectangle_labels_config(root)
        )
        config.validate()
        return config

    @classmethod
    def __parse_data_keys_config(cls, root: ET.Element) -> dict[str, '_DataKey']:
        """Parses the data keys from a Label Studio XML config."""
        config: dict[str, '_DataKey'] = {}
        for element in root:
            if 'value' in element.attrib and element.attrib['value'][0] == '$':
                external_col_name = element.attrib['value'][1:]
                name = element.attrib.get('name')
                column_type = _LS_TAG_MAP.get(element.tag.lower())
                if column_type is None:
                    raise excs.Error(
                        f'Unsupported Label Studio data type: `{element.tag}` (in data key `{external_col_name}`)'
                    )
                config[external_col_name] = _DataKey(name=name, column_type=column_type)
        return config

    @classmethod
    def __parse_rectangle_labels_config(cls, root: ET.Element) -> dict[str, '_RectangleLabel']:
        """Parses the RectangleLabels from a Label Studio XML config."""
        config: dict[str, '_RectangleLabel'] = {}
        for element in root:
            if element.tag.lower() == 'rectanglelabels':
                name = element.attrib['name']
                to_name = element.attrib['toName']
                labels = [child.attrib['value'] for child in element if child.tag.lower() == 'label']
                for label in labels:
                    if label not in coco.COCO_2017_CATEGORIES.values():
                        raise excs.Error(f'Label in `rectanglelabels` config is not a valid COCO object name: {label}')
                config[name] = _RectangleLabel(to_name=to_name, labels=labels)
        return config

    @classmethod
    def __coco_to_predictions(
        cls, coco_annotations: dict[str, Any], from_name: str, rl_info: '_RectangleLabel', task_id: Optional[int] = None
    ) -> dict[str, Any]:
        width = coco_annotations['image']['width']
        height = coco_annotations['image']['height']
        result = [
            {
                'id': f'result_{i}',
                'type': 'rectanglelabels',
                'from_name': from_name,
                'to_name': rl_info.to_name,
                'image_rotation': 0,
                'original_width': width,
                'original_height': height,
                'value': {
                    'rotation': 0,
                    # Label Studio expects image coordinates as % of image dimensions
                    'x': entry['bbox'][0] * 100.0 / width,
                    'y': entry['bbox'][1] * 100.0 / height,
                    'width': entry['bbox'][2] * 100.0 / width,
                    'height': entry['bbox'][3] * 100.0 / height,
                    'rectanglelabels': [coco.COCO_2017_CATEGORIES[entry['category']]],
                },
            }
            for i, entry in enumerate(coco_annotations['annotations'])
            # include only the COCO labels that match a rectanglelabel name
            if coco.COCO_2017_CATEGORIES[entry['category']] in rl_info.labels
        ]
        if task_id is not None:
            return {'task': task_id, 'result': result}
        else:
            return {'result': result}

    def delete(self) -> None:
        """
        Deletes this Label Studio project. This will remove all data and annotations
        associated with this project in Label Studio.
        """
        title = self.project_title
        _label_studio_client().delete_project(self.project_id)
        env.Env.get().console_logger.info(f'Deleted Label Studio project: {title}')

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LabelStudioProject) and self.project_id == other.project_id

    def __hash__(self) -> int:
        return hash(self.project_id)

    @classmethod
    def create(
        cls,
        t: Table,
        label_config: str,
        name: Optional[str],
        title: Optional[str],
        media_import_method: Literal['post', 'file', 'url'],
        col_mapping: Optional[dict[str, str]],
        s3_configuration: Optional[dict[str, Any]],
        **kwargs: Any,
    ) -> 'LabelStudioProject':
        """
        Creates a new Label Studio project, using the Label Studio client configured in Pixeltable.
        """
        # Check that the config is valid before creating the project
        config = cls.__parse_project_config(label_config)

        if name is None:
            # Create a default name that's unique to the table
            all_stores = t.external_stores
            n = 0
            while f'ls_project_{n}' in all_stores:
                n += 1
            name = f'ls_project_{n}'

        if title is None:
            # `title` defaults to table name
            title = t._name

        # Create a column to hold the annotations, if one does not yet exist
        if col_mapping is None or ANNOTATIONS_COLUMN in col_mapping.values():
            if col_mapping is None:
                local_annotations_column = ANNOTATIONS_COLUMN
            else:
                local_annotations_column = next(k for k, v in col_mapping.items() if v == ANNOTATIONS_COLUMN)
            if local_annotations_column not in t._schema:
                t.add_columns({local_annotations_column: pxt.JsonType(nullable=True)})

        resolved_col_mapping = cls.validate_columns(
            t, config.export_columns, {ANNOTATIONS_COLUMN: pxt.JsonType(nullable=True)}, col_mapping
        )

        # Perform some additional validation
        if media_import_method == 'post' and len(config.data_keys) > 1:
            raise excs.Error('`media_import_method` cannot be `post` if there is more than one data key')

        if s3_configuration is not None:
            if media_import_method != 'url':
                raise excs.Error("`s3_configuration` is only valid when `media_import_method == 'url'`")
            s3_configuration = copy.copy(s3_configuration)
            if 'bucket' not in s3_configuration:
                raise excs.Error('`s3_configuration` must contain a `bucket` field')
            if 'title' not in s3_configuration:
                s3_configuration['title'] = 'Pixeltable-S3-Import-Storage'
            if (
                'aws_access_key_id' not in s3_configuration
                and 'aws_secret_access_key' not in s3_configuration
                and 'aws_session_token' not in s3_configuration
            ):
                # Attempt to fill any missing credentials from the environment
                try:
                    import boto3

                    s3_credentials = boto3.Session().get_credentials().get_frozen_credentials()
                    _logger.info(f'Using AWS credentials from the environment for Label Studio project: {title}')
                    s3_configuration['aws_access_key_id'] = s3_credentials.access_key
                    s3_configuration['aws_secret_access_key'] = s3_credentials.secret_key
                    s3_configuration['aws_session_token'] = s3_credentials.token
                except Exception as exc:
                    # This is not necessarily a problem, but we should log that it happened
                    _logger.debug(f'Unable to retrieve AWS credentials from the environment: {exc}')
                    pass

        _logger.info(f'Creating Label Studio project: {title}')
        project = _label_studio_client().start_project(title=title, label_config=label_config, **kwargs)

        if media_import_method == 'file':
            # We need to set up a local storage connection to receive media files
            os.environ['LABEL_STUDIO_LOCAL_FILES_DOCUMENT_ROOT'] = str(Config.get().home)
            try:
                project.connect_local_import_storage(local_store_path=str(env.Env.get().media_dir))
            except HTTPError as exc:
                if exc.errno == 400:
                    response: dict = json.loads(exc.response.text)
                    if (
                        'validation_errors' in response
                        and 'non_field_errors' in response['validation_errors']
                        and 'LOCAL_FILES_SERVING_ENABLED' in response['validation_errors']['non_field_errors'][0]
                    ):
                        raise excs.Error(
                            '`media_import_method` is set to `file`, but your Label Studio server is not configured '
                            'for local file storage.\nPlease set the `LABEL_STUDIO_LOCAL_FILES_SERVING_ENABLED` '
                            'environment variable to `true` in the environment where your Label Studio server '
                            'is running.'
                        ) from exc
                raise  # Handle any other exception type normally

        if s3_configuration is not None:
            _logger.info(f'Setting up S3 import storage for Label Studio project: {title}')
            project.connect_s3_import_storage(**s3_configuration)

        project_id = project.get_params()['id']
        return LabelStudioProject(name, project_id, media_import_method, resolved_col_mapping)


@dataclass(frozen=True)
class _DataKey:
    name: Optional[str]  # The 'name' attribute of the data key; may differ from the field name
    column_type: pxt.ColumnType


@dataclass(frozen=True)
class _RectangleLabel:
    to_name: str
    labels: list[str]


@dataclass(frozen=True)
class _LabelStudioConfig:
    data_keys: dict[str, _DataKey]
    rectangle_labels: dict[str, _RectangleLabel]

    def validate(self) -> None:
        data_key_names = {key.name for key in self.data_keys.values() if key.name is not None}
        for name, rl in self.rectangle_labels.items():
            if rl.to_name not in data_key_names:
                raise excs.Error(
                    f'Invalid Label Studio configuration: `toName` attribute of RectangleLabels `{name}` '
                    f'references an unknown data key: `{rl.to_name}`'
                )

    @property
    def export_columns(self) -> dict[str, pxt.ColumnType]:
        data_key_cols = {key_id: key_info.column_type for key_id, key_info in self.data_keys.items()}
        rl_cols = {name: pxt.JsonType() for name in self.rectangle_labels}
        return {**data_key_cols, **rl_cols}


ANNOTATIONS_COLUMN = 'annotations'
_PAGE_SIZE = 100  # This is the default used in the LS SDK
_LS_TAG_MAP = {
    'header': pxt.StringType(),
    'text': pxt.StringType(),
    'image': pxt.ImageType(),
    'video': pxt.VideoType(),
    'audio': pxt.AudioType(),
}



from __future__ import annotations

import datetime
import enum
from typing import Union

# Python types corresponding to our literal types
LiteralPythonTypes = Union[str, int, float, bool, datetime.datetime]


def print_slice(s: slice) -> str:
    start_str = f'{str(s.start) if s.start is not None else ""}'
    stop_str = f'{str(s.stop) if s.stop is not None else ""}'
    step_str = f'{str(s.step) if s.step is not None else ""}'
    return f'{start_str}:{stop_str}{":" if s.step is not None else ""}{step_str}'


class ComparisonOperator(enum.Enum):
    LT = 0
    LE = 1
    EQ = 2
    NE = 3
    GT = 4
    GE = 5

    def __str__(self) -> str:
        if self == self.LT:
            return '<'
        if self == self.LE:
            return '<='
        if self == self.EQ:
            return '=='
        if self == self.NE:
            return '!='
        if self == self.GT:
            return '>'
        if self == self.GE:
            return '>='
        raise AssertionError()

    def reverse(self) -> ComparisonOperator:
        if self == self.LT:
            return self.GT
        if self == self.LE:
            return self.GE
        if self == self.GT:
            return self.LT
        if self == self.GE:
            return self.LE
        return self


class LogicalOperator(enum.Enum):
    AND = 0
    OR = 1
    NOT = 2

    def __str__(self) -> str:
        if self == self.AND:
            return '&'
        if self == self.OR:
            return '|'
        if self == self.NOT:
            return '~'
        raise AssertionError()


class ArithmeticOperator(enum.Enum):
    ADD = 0
    SUB = 1
    MUL = 2
    DIV = 3
    MOD = 4
    FLOORDIV = 5

    def __str__(self) -> str:
        if self == self.ADD:
            return '+'
        if self == self.SUB:
            return '-'
        if self == self.MUL:
            return '*'
        if self == self.DIV:
            return '/'
        if self == self.MOD:
            return '%'
        if self == self.FLOORDIV:
            return '//'
        raise AssertionError()


class StringOperator(enum.Enum):
    CONCAT = 0
    REPEAT = 1

    def __str__(self) -> str:
        if self == self.CONCAT:
            return '+'
        if self == self.REPEAT:
            return '*'
        raise AssertionError()



import logging
import uuid
from fractions import Fraction
from pathlib import Path
from typing import Any, ClassVar, Optional

import av

from pixeltable import env, exceptions as excs, type_system as ts

from .base import ComponentIterator

_logger = logging.getLogger('pixeltable')


class AudioSplitter(ComponentIterator):
    """
    Iterator over chunks of an audio file. The audio file is split into smaller chunks,
    where the duration of each chunk is determined by chunk_duration_sec.
    The iterator yields audio chunks as pxt.Audio, along with the start and end time of each chunk.
    If the input contains no audio, no chunks are yielded.

    Args:
        chunk_duration_sec: Audio chunk duration in seconds
        overlap_sec: Overlap between consecutive chunks in seconds.
        min_chunk_duration_sec: Drop the last chunk if it is smaller than min_chunk_duration_sec
    """

    # Input parameters
    audio_path: Path
    chunk_duration_sec: float
    overlap_sec: float

    # audio stream details
    container: av.container.input.InputContainer
    audio_time_base: Fraction  # seconds per presentation time

    # List of chunks to extract
    # Each chunk is defined by start and end presentation timestamps in audio file (int)
    chunks_to_extract_in_pts: Optional[list[tuple[int, int]]]
    # next chunk to extract
    next_pos: int

    __codec_map: ClassVar[dict[str, str]] = {
        'mp3': 'mp3',  # MP3 decoder -> mp3/libmp3lame encoder
        'mp3float': 'mp3',  # MP3float decoder -> mp3 encoder
        'aac': 'aac',  # AAC decoder -> AAC encoder
        'vorbis': 'libvorbis',  # Vorbis decoder -> libvorbis encoder
        'opus': 'libopus',  # Opus decoder -> libopus encoder
        'flac': 'flac',  # FLAC decoder -> FLAC encoder
        'wavpack': 'wavpack',  # WavPack decoder -> WavPack encoder
        'alac': 'alac',  # ALAC decoder -> ALAC encoder
    }

    def __init__(
        self, audio: str, chunk_duration_sec: float, *, overlap_sec: float = 0.0, min_chunk_duration_sec: float = 0.0
    ):
        if chunk_duration_sec <= 0.0:
            raise excs.Error('chunk_duration_sec must be a positive number')
        if chunk_duration_sec < min_chunk_duration_sec:
            raise excs.Error('chunk_duration_sec must be at least min_chunk_duration_sec')
        if overlap_sec >= chunk_duration_sec:
            raise excs.Error('overlap_sec must be less than chunk_duration_sec')
        audio_path = Path(audio)
        assert audio_path.exists() and audio_path.is_file()
        self.audio_path = audio_path
        self.next_pos = 0
        self.container = av.open(str(audio_path))
        if len(self.container.streams.audio) == 0:
            # No audio stream
            return
        self.chunk_duration_sec = chunk_duration_sec
        self.overlap_sec = overlap_sec
        self.min_chunk_duration_sec = min_chunk_duration_sec
        self.audio_time_base = self.container.streams.audio[0].time_base

        audio_start_time_pts = self.container.streams.audio[0].start_time or 0
        audio_start_time_sec = float(audio_start_time_pts * self.audio_time_base)
        total_audio_duration_pts = self.container.streams.audio[0].duration or 0
        total_audio_duration_sec = float(total_audio_duration_pts * self.audio_time_base)

        self.chunks_to_extract_in_pts = [
            (round(start / self.audio_time_base), round(end / self.audio_time_base))
            for (start, end) in self.build_chunks(
                audio_start_time_sec, total_audio_duration_sec, chunk_duration_sec, overlap_sec, min_chunk_duration_sec
            )
        ]
        _logger.debug(
            f'AudioIterator: path={self.audio_path} total_audio_duration_pts={total_audio_duration_pts} '
            f'chunks_to_extract_in_pts={self.chunks_to_extract_in_pts}'
        )

    @classmethod
    def build_chunks(
        cls,
        start_time_sec: float,
        total_duration_sec: float,
        chunk_duration_sec: float,
        overlap_sec: float,
        min_chunk_duration_sec: float,
    ) -> list[tuple[float, float]]:
        chunks_to_extract_in_sec: list[tuple[float, float]] = []
        current_pos = start_time_sec
        end_time = start_time_sec + total_duration_sec
        while current_pos < end_time:
            chunk_start = current_pos
            chunk_end = min(chunk_start + chunk_duration_sec, end_time)
            chunks_to_extract_in_sec.append((chunk_start, chunk_end))
            if chunk_end >= end_time:
                break
            current_pos = chunk_end - overlap_sec
        # If the last chunk is smaller than min_chunk_duration_sec then drop the last chunk from the list
        if (
            len(chunks_to_extract_in_sec) > 0
            and (chunks_to_extract_in_sec[-1][1] - chunks_to_extract_in_sec[-1][0]) < min_chunk_duration_sec
        ):
            return chunks_to_extract_in_sec[:-1]  # return all but the last chunk
        return chunks_to_extract_in_sec

    @classmethod
    def input_schema(cls) -> dict[str, ts.ColumnType]:
        return {
            'audio': ts.AudioType(nullable=False),
            'chunk_duration_sec': ts.FloatType(nullable=True),
            'overlap_sec': ts.FloatType(nullable=True),
            'min_chunk_duration_sec': ts.FloatType(nullable=True),
        }

    @classmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> tuple[dict[str, ts.ColumnType], list[str]]:
        return {
            'start_time_sec': ts.FloatType(),
            'end_time_sec': ts.FloatType(),
            'audio_chunk': ts.AudioType(nullable=True),
        }, []

    def __next__(self) -> dict[str, Any]:
        if self.next_pos >= len(self.chunks_to_extract_in_pts):
            raise StopIteration
        target_chunk_start, target_chunk_end = self.chunks_to_extract_in_pts[self.next_pos]
        chunk_start_pts = 0
        chunk_end_pts = 0
        chunk_file = str(env.Env.get().tmp_dir / f'{uuid.uuid4()}{self.audio_path.suffix}')
        output_container = av.open(chunk_file, mode='w')
        input_stream = self.container.streams.audio[0]
        codec_name = AudioSplitter.__codec_map.get(input_stream.codec_context.name, input_stream.codec_context.name)
        output_stream = output_container.add_stream(codec_name, rate=input_stream.codec_context.sample_rate)
        assert isinstance(output_stream, av.audio.stream.AudioStream)
        frame_count = 0
        # Since frames don't align with chunk boundaries, we may have read an extra frame in previous iteration
        # Seek to the nearest frame in stream at current chunk start time
        self.container.seek(target_chunk_start, backward=True, stream=self.container.streams.audio[0])
        while True:
            try:
                frame = next(self.container.decode(audio=0))
            except EOFError as e:
                raise excs.Error(f"Failed to read audio file '{self.audio_path}': {e}") from e
            except StopIteration:
                # no more frames to scan
                break
            if frame.pts < target_chunk_start:
                # Current frame is behind chunk's start time, always get frame next to chunk's start time
                continue
            if frame.pts >= target_chunk_end:
                # Frame has crossed the chunk boundary, it should be picked up by next chunk, throw away
                # the current frame
                break
            frame_end = frame.pts + frame.samples
            if frame_count == 0:
                # Record start of the first frame
                chunk_start_pts = frame.pts
            # Write frame to output container
            frame_count += 1
            # If encode returns packets, write them to output container. Some encoders will buffer the frames.
            output_container.mux(output_stream.encode(frame))
            # record this frame's end as chunks end
            chunk_end_pts = frame_end
            # Check if frame's end has crossed the chunk boundary
            if frame_end >= target_chunk_end:
                break

        # record result
        if frame_count > 0:
            # flush encoder
            output_container.mux(output_stream.encode(None))
            output_container.close()
            result = {
                'start_time_sec': round(float(chunk_start_pts * self.audio_time_base), 4),
                'end_time_sec': round(float(chunk_end_pts * self.audio_time_base), 4),
                'audio_chunk': chunk_file if frame_count > 0 else None,
            }
            _logger.debug('audio chunk result: %s', result)
            self.next_pos += 1
            return result
        else:
            # It's possible that there are no frames in the range of the last chunk, stop the iterator in this case.
            # Note that start_time points at the first frame so case applies only for the last chunk
            assert self.next_pos == len(self.chunks_to_extract_in_pts) - 1
            self.next_pos += 1
            raise StopIteration

    def close(self) -> None:
        self.container.close()

    def set_pos(self, pos: int) -> None:
        pass



from typing import AsyncIterator

from pixeltable import catalog, exceptions as excs, exprs

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode


class ComponentIterationNode(ExecNode):
    """Expands each row from a base table into one row per component returned by an iterator

    Returns row batches of OUTPUT_BATCH_SIZE size.
    """

    view: catalog.TableVersionHandle

    __OUTPUT_BATCH_SIZE = 1024

    def __init__(self, view: catalog.TableVersionHandle, input: ExecNode):
        assert view.get().is_component_view
        super().__init__(input.row_builder, [], [], input)
        self.view = view
        iterator_args = [view.get().iterator_args.copy()]
        self.row_builder.set_slot_idxs(iterator_args)
        self.iterator_args = iterator_args[0]
        assert isinstance(self.iterator_args, exprs.InlineDict)
        self.iterator_args_ctx = self.row_builder.create_eval_ctx([self.iterator_args])
        self.iterator_output_schema, self.unstored_column_names = self.view.get().iterator_cls.output_schema(
            **self.iterator_args.to_kwargs()
        )
        self.iterator_output_fields = list(self.iterator_output_schema.keys())
        self.iterator_output_cols = {
            field_name: self.view.get().cols_by_name[field_name] for field_name in self.iterator_output_fields
        }
        # referenced iterator output fields
        self.refd_output_slot_idxs = {
            e.col.name: e.slot_idx
            for e in self.row_builder.unique_exprs
            if isinstance(e, exprs.ColumnRef) and e.col.name in self.iterator_output_fields
        }

    async def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        output_batch = DataRowBatch(self.view, self.row_builder)
        async for input_batch in self.input:
            for input_row in input_batch:
                self.row_builder.eval(input_row, self.iterator_args_ctx)
                iterator_args = input_row[self.iterator_args.slot_idx]
                assert isinstance(iterator_args, dict)
                # We need to ensure that all of the required (non-nullable) parameters of the iterator are
                # specified and are not null. If any of them are null, then we skip this row (i.e., we emit 0
                # output rows for this input row).
                if self.__non_nullable_args_specified(iterator_args):
                    iterator = self.view.get().iterator_cls(**iterator_args)
                    for pos, component_dict in enumerate(iterator):
                        output_row = output_batch.add_row()
                        input_row.copy(output_row)
                        # we're expanding the input and need to add the iterator position to the pk
                        self.__populate_output_row(output_row, pos, component_dict)
                        if len(output_batch) == self.__OUTPUT_BATCH_SIZE:
                            yield output_batch
                            output_batch = DataRowBatch(self.view, self.row_builder)

        if len(output_batch) > 0:
            yield output_batch

    def __non_nullable_args_specified(self, iterator_args: dict) -> bool:
        """
        Returns true if all non-nullable iterator arguments are not `None`.
        """
        input_schema = self.view.get().iterator_cls.input_schema()
        for arg_name, arg_value in iterator_args.items():
            col_type = input_schema[arg_name]
            if arg_value is None and not col_type.nullable:
                return False
        return True

    def __populate_output_row(self, output_row: exprs.DataRow, pos: int, component_dict: dict) -> None:
        pk = output_row.pk[:-1] + (pos,) + output_row.pk[-1:]
        output_row.set_pk(pk)
        # verify and copy component_dict fields to their respective slots in output_row
        for field_name, field_val in component_dict.items():
            if field_name not in self.iterator_output_fields:
                raise excs.Error(
                    f'Invalid field name {field_name} in output of {self.view.get().iterator_cls.__name__}'
                )
            if field_name not in self.refd_output_slot_idxs:
                # we can ignore this
                continue
            output_col = self.iterator_output_cols[field_name]
            output_col.col_type.validate_literal(field_val)
            output_row[self.refd_output_slot_idxs[field_name]] = field_val
        if len(component_dict) != len(self.iterator_output_fields):
            missing_fields = set(self.refd_output_slot_idxs.keys()) - set(component_dict.keys())
            raise excs.Error(
                f'Invalid output of {self.view.get().iterator_cls.__name__}: missing fields {", ".join(missing_fields)}'
            )



import abc

from sqlalchemy import URL


class Dbms(abc.ABC):
    """
    Provides abstractions for utilities to interact with a database system.
    """

    name: str
    transaction_isolation_level: str
    version_index_type: str
    db_url: URL

    def __init__(self, name: str, transaction_isolation_level: str, version_index_type: str, db_url: URL) -> None:
        self.name = name
        self.transaction_isolation_level = transaction_isolation_level
        self.version_index_type = version_index_type
        self.db_url = db_url

    @abc.abstractmethod
    def drop_db_stmt(self, database: str) -> str: ...

    @abc.abstractmethod
    def create_db_stmt(self, database: str) -> str: ...

    @abc.abstractmethod
    def default_system_db_url(self) -> str: ...


class PostgresqlDbms(Dbms):
    """
    Implements utilities to interact with Postgres database.
    """

    def __init__(self, db_url: URL):
        super().__init__('postgresql', 'REPEATABLE READ', 'brin', db_url)

    def drop_db_stmt(self, database: str) -> str:
        return f'DROP DATABASE {database}'

    def create_db_stmt(self, database: str) -> str:
        return f"CREATE DATABASE {database} ENCODING 'utf-8' LC_COLLATE 'C' LC_CTYPE 'C' TEMPLATE template0"

    def default_system_db_url(self) -> str:
        a = self.db_url.set(database='postgres').render_as_string(hide_password=False)
        return a


class CockroachDbms(Dbms):
    """
    Implements utilities to interact with CockroachDb database.
    """

    def __init__(self, db_url: URL):
        super().__init__('cockroachdb', 'SERIALIZABLE', 'btree', db_url)

    def drop_db_stmt(self, database: str) -> str:
        return f'DROP DATABASE {database} CASCADE'

    def create_db_stmt(self, database: str) -> str:
        return f"CREATE DATABASE {database} TEMPLATE template0 ENCODING 'utf-8' LC_COLLATE 'C' LC_CTYPE 'C'"

    def default_system_db_url(self) -> str:
        return self.db_url.set(database='defaultdb').render_as_string(hide_password=False)



import datetime
from typing import Any, Optional

import sqlalchemy as sql

import pixeltable as pxt
from pixeltable.metadata import register_converter, schema
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=19)
def _(engine: sql.engine.Engine) -> None:
    # Convert all timestamp literals to aware datetimes
    convert_table_md(engine, substitution_fn=__update_timestamp_literals)

    # Convert all timestamp columns to TIMESTAMPTZ. (This conversion will take place in the database
    # default time zone, which is what we want, since in versions <= 19 they were naive timestamps.)
    with engine.begin() as conn:
        tables = conn.execute(sql.select(schema.Table.id, schema.Table.md))
        for id, md in tables:
            store_prefix = 'view' if md['view_md'] is not None else 'tbl'
            store_name = f'{store_prefix}_{id.hex}'
            column_md = md['column_md']
            timestamp_cols = [
                col_id for col_id, col in column_md.items() if col['col_type']['_classname'] == 'TimestampType'
            ]
            for col_id in timestamp_cols:
                conn.execute(sql.text(f'ALTER TABLE {store_name} ALTER COLUMN col_{col_id} TYPE TIMESTAMPTZ'))


def __update_timestamp_literals(k: Any, v: Any) -> Optional[tuple[Any, Any]]:
    if isinstance(v, dict) and 'val_t' in v:
        # It's a literal with an explicit 'val_t' field. In version 19 this can only mean a
        # timestamp literal, which (in version 19) is stored in the DB as a naive datetime.
        # We convert it to an aware datetime, stored in UTC.
        assert v['_classname'] == 'Literal'
        assert v['val_t'] == pxt.ColumnType.Type.TIMESTAMP.name
        assert isinstance(v['val'], str)
        dt = datetime.datetime.fromisoformat(v['val'])
        assert dt.tzinfo is None  # In version 19 all timestamps are naive
        dt_utc = dt.astimezone(datetime.timezone.utc)
        v['val'] = dt_utc.isoformat()
        return k, v
    return None



# ruff: noqa: F401

from .arithmetic_expr import ArithmeticExpr
from .array_slice import ArraySlice
from .column_property_ref import ColumnPropertyRef
from .column_ref import ColumnRef
from .comparison import Comparison
from .compound_predicate import CompoundPredicate
from .data_row import DataRow
from .expr import Expr
from .expr_dict import ExprDict
from .expr_set import ExprSet
from .function_call import FunctionCall
from .globals import ArithmeticOperator, ComparisonOperator, LogicalOperator
from .in_predicate import InPredicate
from .inline_expr import InlineArray, InlineDict, InlineList
from .is_null import IsNull
from .json_mapper import JsonMapper, JsonMapperDispatch
from .json_path import JsonPath
from .literal import Literal
from .method_ref import MethodRef
from .object_ref import ObjectRef
from .row_builder import ColumnSlotIdx, ExecProfile, RowBuilder
from .rowid_ref import RowidRef
from .similarity_expr import SimilarityExpr
from .sql_element_cache import SqlElementCache
from .string_op import StringOp
from .type_cast import TypeCast
from .variable import Variable



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=25)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    if k == 'path' and (
        v in ('pixeltable.functions.huggingface.clip_text', 'pixeltable.functions.huggingface.clip_image')
    ):
        return 'path', 'pixeltable.functions.huggingface.clip'
    return None



import logging
from typing import TextIO


def map_level(verbosity: int) -> int:
    """
        Map verbosity level to logging level.
        0 - minimum logging - warn and above
        1 - default logging - info and above
        2 - more logging - debug and above
    Args:

    Returns:
        Logging level as integer
    """
    if verbosity == 0:
        return logging.WARN
    if verbosity == 1:
        return logging.INFO
    if verbosity == 2:
        return logging.DEBUG
    return logging.INFO


class ConsoleOutputHandler(logging.StreamHandler):
    def __init__(self, stream: TextIO):
        super().__init__(stream)

    def emit(self, record: logging.LogRecord) -> None:
        if record.msg.endswith('\n'):
            self.stream.write(record.msg)
        else:
            self.stream.write(record.msg + '\n')


class ConsoleMessageFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return getattr(record, 'user_visible', False)


class ConsoleLogger(logging.LoggerAdapter):
    def __init__(self, logger: logging.Logger):
        super().__init__(logger, extra={'user_visible': True})



from __future__ import annotations

import enum
import json
import logging
import math
import urllib.parse
import urllib.request
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Literal, Optional, Union, cast

import pandas as pd
from pyarrow.parquet import ParquetDataset

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable.io.pandas import _df_check_primary_key_values, _df_row_to_pxt_row, df_infer_schema
from pixeltable.utils import parse_local_file_path

from .utils import normalize_schema_names

_logger = logging.getLogger('pixeltable')

# ---------------------------------------------------------------------------------------------------------

if TYPE_CHECKING:
    import datasets  # type: ignore[import-untyped]

    from pixeltable.globals import RowData, TableDataSource


class TableDataConduitFormat(str, enum.Enum):
    """Supported formats for TableDataConduit"""

    JSON = 'json'
    CSV = 'csv'
    EXCEL = 'excel'
    PARQUET = 'parquet'

    @classmethod
    def is_valid(cls, x: Any) -> bool:
        if isinstance(x, str):
            return x.lower() in [c.value for c in cls]
        return False


# ---------------------------------------------------------------------------------------------------------


@dataclass
class TableDataConduit:
    source: TableDataSource
    source_format: Optional[str] = None
    source_column_map: Optional[dict[str, str]] = None
    if_row_exists: Literal['update', 'ignore', 'error'] = 'error'
    pxt_schema: Optional[dict[str, Any]] = None
    src_schema_overrides: Optional[dict[str, Any]] = None
    src_schema: Optional[dict[str, Any]] = None
    pxt_pk: Optional[list[str]] = None
    src_pk: Optional[list[str]] = None
    valid_rows: Optional[RowData] = None
    extra_fields: dict[str, Any] = field(default_factory=dict)

    reqd_col_names: set[str] = field(default_factory=set)
    computed_col_names: set[str] = field(default_factory=set)

    total_rows: int = 0  # total number of rows emitted via valid_row_batch Iterator

    _K_BATCH_SIZE_BYTES = 100_000_000  # 100 MB

    def check_source_format(self) -> None:
        assert self.source_format is None or TableDataConduitFormat.is_valid(self.source_format)

    @classmethod
    def is_rowdata_structure(cls, d: TableDataSource) -> bool:
        if not isinstance(d, list) or len(d) == 0:
            return False
        return all(isinstance(row, dict) for row in d)

    def is_direct_df(self) -> bool:
        return isinstance(self.source, pxt.DataFrame) and self.source_column_map is None

    def normalize_pxt_schema_types(self) -> None:
        for name, coltype in self.pxt_schema.items():
            self.pxt_schema[name] = pxt.ColumnType.normalize_type(coltype)

    def infer_schema(self) -> dict[str, Any]:
        raise NotImplementedError

    def valid_row_batch(self) -> Iterator[RowData]:
        raise NotImplementedError

    def prepare_for_insert_into_table(self) -> None:
        if self.source is None:
            return
        raise NotImplementedError

    def add_table_info(self, table: pxt.Table) -> None:
        """Add information about the table into which we are inserting data"""
        assert isinstance(table, pxt.Table)
        self.pxt_schema = table._schema
        self.pxt_pk = table._tbl_version.get().primary_key
        for col in table._tbl_version_path.columns():
            if col.is_required_for_insert:
                self.reqd_col_names.add(col.name)
            if col.is_computed:
                self.computed_col_names.add(col.name)
        self.src_pk = []

    # Check source columns : required, computed, unknown
    def check_source_columns_are_insertable(self, columns: Iterable[str]) -> None:
        col_name_set: set[str] = set()
        for col_name in columns:  # FIXME
            mapped_col_name = self.source_column_map.get(col_name, col_name)
            col_name_set.add(mapped_col_name)
            if mapped_col_name not in self.pxt_schema:
                raise excs.Error(f'Unknown column name {mapped_col_name}')
            if mapped_col_name in self.computed_col_names:
                raise excs.Error(f'Value for computed column {mapped_col_name}')
        missing_cols = self.reqd_col_names - col_name_set
        if len(missing_cols) > 0:
            raise excs.Error(f'Missing required column(s) ({", ".join(missing_cols)})')


# ---------------------------------------------------------------------------------------------------------


class DFTableDataConduit(TableDataConduit):
    pxt_df: pxt.DataFrame = None

    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> 'DFTableDataConduit':
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        assert isinstance(tds.source, pxt.DataFrame)
        t.pxt_df = tds.source
        return t

    def infer_schema(self) -> dict[str, Any]:
        self.pxt_schema = self.pxt_df.schema
        self.pxt_pk = self.src_pk
        return self.pxt_schema

    def prepare_for_insert_into_table(self) -> None:
        if self.source_column_map is None:
            self.source_column_map = {}
        self.check_source_columns_are_insertable(self.pxt_df.schema.keys())


# ---------------------------------------------------------------------------------------------------------


class RowDataTableDataConduit(TableDataConduit):
    raw_rows: Optional[RowData] = None
    disable_mapping: bool = True
    batch_count: int = 0

    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> 'RowDataTableDataConduit':
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        if isinstance(tds.source, Iterator):
            # Instantiate the iterator to get the raw rows here
            t.raw_rows = list(tds.source)
        elif TYPE_CHECKING:
            t.raw_rows = cast(RowData, tds.source)
        else:
            t.raw_rows = tds.source
        t.batch_count = 0
        return t

    def infer_schema(self) -> dict[str, Any]:
        from .datarows import _infer_schema_from_rows

        if self.source_column_map is None:
            if self.src_schema_overrides is None:
                self.src_schema_overrides = {}
            self.src_schema = _infer_schema_from_rows(self.raw_rows, self.src_schema_overrides, self.src_pk)
            self.pxt_schema, self.pxt_pk, self.source_column_map = normalize_schema_names(
                self.src_schema, self.src_pk, self.src_schema_overrides, self.disable_mapping
            )
            self.normalize_pxt_schema_types()
        else:
            raise NotImplementedError()

        self.prepare_for_insert_into_table()
        return self.pxt_schema

    def prepare_for_insert_into_table(self) -> None:
        # Converting rows to insertable format is not needed, misnamed columns and types
        # are errors in the incoming row format
        if self.source_column_map is None:
            self.source_column_map = {}
        self.valid_rows = [self._translate_row(row) for row in self.raw_rows]

        self.batch_count = 1 if self.raw_rows is not None else 0

    def _translate_row(self, row: dict[str, Any]) -> dict[str, Any]:
        if not isinstance(row, dict):
            raise excs.Error(f'row {row} is not a dictionary')

        col_names: set[str] = set()
        output_row: dict[str, Any] = {}
        for col_name, val in row.items():
            mapped_col_name = self.source_column_map.get(col_name, col_name)
            col_names.add(mapped_col_name)
            if mapped_col_name not in self.pxt_schema:
                raise excs.Error(f'Unknown column name {mapped_col_name} in row {row}')
            if mapped_col_name in self.computed_col_names:
                raise excs.Error(f'Value for computed column {mapped_col_name} in row {row}')
            # basic sanity checks here
            try:
                checked_val = self.pxt_schema[mapped_col_name].create_literal(val)
            except TypeError as e:
                msg = str(e)
                raise excs.Error(f'Error in column {col_name}: {msg[0].lower() + msg[1:]}\nRow: {row}') from e
            output_row[mapped_col_name] = checked_val
        missing_cols = self.reqd_col_names - col_names
        if len(missing_cols) > 0:
            raise excs.Error(f'Missing required column(s) ({", ".join(missing_cols)}) in row {row}')
        return output_row

    def valid_row_batch(self) -> Iterator[RowData]:
        if self.batch_count > 0:
            self.batch_count -= 1
            yield self.valid_rows


# ---------------------------------------------------------------------------------------------------------


class PandasTableDataConduit(TableDataConduit):
    pd_df: pd.DataFrame = None
    batch_count: int = 0

    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> PandasTableDataConduit:
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        assert isinstance(tds.source, pd.DataFrame)
        t.pd_df = tds.source
        t.batch_count = 0
        return t

    def infer_schema_part1(self) -> tuple[dict[str, Any], list[str]]:
        """Return inferred schema, inferred primary key, and source column map"""
        if self.source_column_map is None:
            if self.src_schema_overrides is None:
                self.src_schema_overrides = {}
            self.src_schema = df_infer_schema(self.pd_df, self.src_schema_overrides, self.src_pk)
            inferred_schema, inferred_pk, self.source_column_map = normalize_schema_names(
                self.src_schema, self.src_pk, self.src_schema_overrides, False
            )
            return inferred_schema, inferred_pk
        else:
            raise NotImplementedError()

    def infer_schema(self) -> dict[str, Any]:
        self.pxt_schema, self.pxt_pk = self.infer_schema_part1()
        self.normalize_pxt_schema_types()
        _df_check_primary_key_values(self.pd_df, self.src_pk)
        self.prepare_insert()
        return self.pxt_schema

    def prepare_for_insert_into_table(self) -> None:
        _, inferred_pk = self.infer_schema_part1()
        assert len(inferred_pk) == 0
        self.prepare_insert()

    def prepare_insert(self) -> None:
        if self.source_column_map is None:
            self.source_column_map = {}
        self.check_source_columns_are_insertable(self.pd_df.columns)
        # Convert all rows to insertable format
        self.valid_rows = [
            _df_row_to_pxt_row(row, self.src_schema, self.source_column_map) for row in self.pd_df.itertuples()
        ]
        self.batch_count = 1

    def valid_row_batch(self) -> Iterator[RowData]:
        if self.batch_count > 0:
            self.batch_count -= 1
            yield self.valid_rows


# ---------------------------------------------------------------------------------------------------------


class CSVTableDataConduit(TableDataConduit):
    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> 'PandasTableDataConduit':
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        assert isinstance(t.source, str)
        t.source = pd.read_csv(t.source, **t.extra_fields)
        return PandasTableDataConduit.from_tds(t)


# ---------------------------------------------------------------------------------------------------------


class ExcelTableDataConduit(TableDataConduit):
    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> 'PandasTableDataConduit':
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        assert isinstance(t.source, str)
        t.source = pd.read_excel(t.source, **t.extra_fields)
        return PandasTableDataConduit.from_tds(t)


# ---------------------------------------------------------------------------------------------------------


class JsonTableDataConduit(TableDataConduit):
    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> RowDataTableDataConduit:
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        assert isinstance(t.source, str)

        path = parse_local_file_path(t.source)
        if path is None:  # it's a URL
            # TODO: This should read from S3 as well.
            contents = urllib.request.urlopen(t.source).read()
        else:
            with open(path, 'r', encoding='utf-8') as fp:
                contents = fp.read()
        rows = json.loads(contents, **t.extra_fields)
        t.source = rows
        t2 = RowDataTableDataConduit.from_tds(t)
        t2.disable_mapping = False
        return t2


# ---------------------------------------------------------------------------------------------------------


class HFTableDataConduit(TableDataConduit):
    hf_ds: Optional[Union[datasets.Dataset, datasets.DatasetDict]] = None
    column_name_for_split: Optional[str] = None
    categorical_features: dict[str, dict[int, str]]
    hf_schema: dict[str, Any] = None
    dataset_dict: dict[str, datasets.Dataset] = None
    hf_schema_source: dict[str, Any] = None

    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> 'HFTableDataConduit':
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)
        import datasets

        assert isinstance(tds.source, (datasets.Dataset, datasets.DatasetDict))
        t.hf_ds = tds.source
        if 'column_name_for_split' in t.extra_fields:
            t.column_name_for_split = t.extra_fields['column_name_for_split']
        return t

    @classmethod
    def is_applicable(cls, tds: TableDataConduit) -> bool:
        try:
            import datasets

            return (isinstance(tds.source_format, str) and tds.source_format.lower() == 'huggingface') or isinstance(
                tds.source, (datasets.Dataset, datasets.DatasetDict)
            )
        except ImportError:
            return False

    def infer_schema_part1(self) -> tuple[dict[str, Any], list[str]]:
        from pixeltable.io.hf_datasets import _get_hf_schema, huggingface_schema_to_pxt_schema

        if self.source_column_map is None:
            if self.src_schema_overrides is None:
                self.src_schema_overrides = {}
            self.hf_schema_source = _get_hf_schema(self.hf_ds)
            self.src_schema = huggingface_schema_to_pxt_schema(
                self.hf_schema_source, self.src_schema_overrides, self.src_pk
            )

            # Add the split column to the schema if requested
            if self.column_name_for_split is not None:
                if self.column_name_for_split in self.src_schema:
                    raise excs.Error(
                        f'Column name `{self.column_name_for_split}` already exists in dataset schema;'
                        f'provide a different `column_name_for_split`'
                    )
                self.src_schema[self.column_name_for_split] = pxt.StringType(nullable=True)

            inferred_schema, inferred_pk, self.source_column_map = normalize_schema_names(
                self.src_schema, self.src_pk, self.src_schema_overrides, True
            )
            return inferred_schema, inferred_pk
        else:
            raise NotImplementedError()

    def infer_schema(self) -> dict[str, Any]:
        self.pxt_schema, self.pxt_pk = self.infer_schema_part1()
        self.normalize_pxt_schema_types()
        self.prepare_insert()
        return self.pxt_schema

    def prepare_for_insert_into_table(self) -> None:
        _, inferred_pk = self.infer_schema_part1()
        assert len(inferred_pk) == 0
        self.prepare_insert()

    def prepare_insert(self) -> None:
        import datasets

        if isinstance(self.source, datasets.Dataset):
            # when loading an hf dataset partially, dataset.split._name is sometimes the form "train[0:1000]"
            raw_name = self.source.split._name
            split_name = raw_name.split('[')[0] if raw_name is not None else None
            self.dataset_dict = {split_name: self.source}
        else:
            assert isinstance(self.source, datasets.DatasetDict)
            self.dataset_dict = self.source

        # extract all class labels from the dataset to translate category ints to strings
        self.categorical_features = {
            feature_name: feature_type.names
            for (feature_name, feature_type) in self.hf_schema_source.items()
            if isinstance(feature_type, datasets.ClassLabel)
        }
        if self.source_column_map is None:
            self.source_column_map = {}
        self.check_source_columns_are_insertable(self.hf_schema_source.keys())

    def _translate_row(self, row: dict[str, Any], split_name: str) -> dict[str, Any]:
        output_row: dict[str, Any] = {}
        for col_name, val in row.items():
            # translate category ints to strings
            new_val = self.categorical_features[col_name][val] if col_name in self.categorical_features else val
            mapped_col_name = self.source_column_map.get(col_name, col_name)

            # Convert values to the appropriate type if needed
            try:
                checked_val = self.pxt_schema[mapped_col_name].create_literal(new_val)
            except TypeError as e:
                msg = str(e)
                raise excs.Error(f'Error in column {col_name}: {msg[0].lower() + msg[1:]}\nRow: {row}') from e
            output_row[mapped_col_name] = checked_val

        # add split name to output row
        if self.column_name_for_split is not None:
            output_row[self.column_name_for_split] = split_name
        return output_row

    def valid_row_batch(self) -> Iterator[RowData]:
        for split_name, split_dataset in self.dataset_dict.items():
            num_batches = split_dataset.size_in_bytes / self._K_BATCH_SIZE_BYTES
            tuples_per_batch = math.ceil(split_dataset.num_rows / num_batches)
            assert tuples_per_batch > 0

            batch = []
            for row in split_dataset:
                batch.append(self._translate_row(row, split_name))
                if len(batch) >= tuples_per_batch:
                    yield batch
                    batch = []
            # last batch
            if len(batch) > 0:
                yield batch


# ---------------------------------------------------------------------------------------------------------


class ParquetTableDataConduit(TableDataConduit):
    pq_ds: Optional[ParquetDataset] = None

    @classmethod
    def from_tds(cls, tds: TableDataConduit) -> 'ParquetTableDataConduit':
        tds_fields = {f.name for f in fields(tds)}
        kwargs = {k: v for k, v in tds.__dict__.items() if k in tds_fields}
        t = cls(**kwargs)

        from pyarrow import parquet

        assert isinstance(tds.source, str)
        input_path = Path(tds.source).expanduser()
        t.pq_ds = parquet.ParquetDataset(str(input_path))
        return t

    def infer_schema_part1(self) -> tuple[dict[str, Any], list[str]]:
        from pixeltable.utils.arrow import ar_infer_schema

        if self.source_column_map is None:
            if self.src_schema_overrides is None:
                self.src_schema_overrides = {}
            self.src_schema = ar_infer_schema(self.pq_ds.schema, self.src_schema_overrides, self.src_pk)
            inferred_schema, inferred_pk, self.source_column_map = normalize_schema_names(
                self.src_schema, self.src_pk, self.src_schema_overrides
            )
            return inferred_schema, inferred_pk
        else:
            raise NotImplementedError()

    def infer_schema(self) -> dict[str, Any]:
        self.pxt_schema, self.pxt_pk = self.infer_schema_part1()
        self.normalize_pxt_schema_types()
        self.prepare_insert()
        return self.pxt_schema

    def prepare_for_insert_into_table(self) -> None:
        _, inferred_pk = self.infer_schema_part1()
        assert len(inferred_pk) == 0
        self.prepare_insert()

    def prepare_insert(self) -> None:
        if self.source_column_map is None:
            self.source_column_map = {}
        self.check_source_columns_are_insertable(self.pq_ds.schema.names)
        self.total_rows = 0

    def valid_row_batch(self) -> Iterator[RowData]:
        from pixeltable.utils.arrow import iter_tuples2

        try:
            for fragment in self.pq_ds.fragments:  # type: ignore[attr-defined]
                for batch in fragment.to_batches():
                    dict_batch = list(iter_tuples2(batch, self.source_column_map, self.pxt_schema))
                    self.total_rows += len(dict_batch)
                    yield dict_batch
        except Exception as e:
            _logger.error(f'Error after inserting {self.total_rows} rows from Parquet file into table: {e}')
            raise e


# ---------------------------------------------------------------------------------------------------------


class UnkTableDataConduit(TableDataConduit):
    """Source type is not known at the time of creation"""

    def specialize(self) -> TableDataConduit:
        if isinstance(self.source, pxt.DataFrame):
            return DFTableDataConduit.from_tds(self)
        if isinstance(self.source, pd.DataFrame):
            return PandasTableDataConduit.from_tds(self)
        if HFTableDataConduit.is_applicable(self):
            return HFTableDataConduit.from_tds(self)
        if self.source_format == 'csv' or (isinstance(self.source, str) and '.csv' in self.source.lower()):
            return CSVTableDataConduit.from_tds(self)
        if self.source_format == 'excel' or (isinstance(self.source, str) and '.xls' in self.source.lower()):
            return ExcelTableDataConduit.from_tds(self)
        if self.source_format == 'json' or (isinstance(self.source, str) and '.json' in self.source.lower()):
            return JsonTableDataConduit.from_tds(self)
        if self.source_format == 'parquet' or (
            isinstance(self.source, str) and any(s in self.source.lower() for s in ['.parquet', '.pq', '.parq'])
        ):
            return ParquetTableDataConduit.from_tds(self)
        if (
            self.is_rowdata_structure(self.source)
            # An Iterator as a source is assumed to produce rows
            or isinstance(self.source, Iterator)
        ):
            return RowDataTableDataConduit.from_tds(self)

        raise excs.Error(f'Unsupported data source type: {type(self.source)}')



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=22)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    if isinstance(v, dict) and '_classname' in v and v['_classname'] == 'DataFrame':
        v['from_clause'] = {'tbls': [v['tbl']], 'join_clauses': []}
        return k, v
    return None



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `VideoType`.

Example:
```python
import pixeltable as pxt
from pixeltable.functions import video as pxt_video

t = pxt.get_table(...)
t.select(pxt_video.extract_audio(t.video_col)).collect()
```
"""

import tempfile
import uuid
from pathlib import Path
from typing import Any, Optional

import av
import numpy as np
import PIL.Image

import pixeltable as pxt
from pixeltable import env
from pixeltable.utils.code import local_public_names

_format_defaults = {  # format -> (codec, ext)
    'wav': ('pcm_s16le', 'wav'),
    'mp3': ('libmp3lame', 'mp3'),
    'flac': ('flac', 'flac'),
    # 'mp4': ('aac', 'm4a'),
}

# for mp4:
# - extract_audio() fails with
#   "Application provided invalid, non monotonically increasing dts to muxer in stream 0: 1146 >= 290"
# - chatgpt suggests this can be fixed in the following manner
#     for packet in container.demux(audio_stream):
#         packet.pts = None  # Reset the PTS and DTS to allow FFmpeg to set them automatically
#         packet.dts = None
#         for frame in packet.decode():
#             frame.pts = None
#             for packet in output_stream.encode(frame):
#                 output_container.mux(packet)
#
#     # Flush remaining packets
#     for packet in output_stream.encode():
#         output_container.mux(packet)


@pxt.uda(requires_order_by=True)
class make_video(pxt.Aggregator):
    """
    Aggregator that creates a video from a sequence of images.
    """

    container: Optional[av.container.OutputContainer]
    stream: Optional[av.video.stream.VideoStream]
    fps: int

    def __init__(self, fps: int = 25):
        """follows https://pyav.org/docs/develop/cookbook/numpy.html#generating-video"""
        self.container = None
        self.stream = None
        self.fps = fps

    def update(self, frame: PIL.Image.Image) -> None:
        if frame is None:
            return
        if self.container is None:
            (_, output_filename) = tempfile.mkstemp(suffix='.mp4', dir=str(env.Env.get().tmp_dir))
            self.out_file = Path(output_filename)
            self.container = av.open(str(self.out_file), mode='w')
            self.stream = self.container.add_stream('h264', rate=self.fps)
            self.stream.pix_fmt = 'yuv420p'
            self.stream.width = frame.width
            self.stream.height = frame.height

        av_frame = av.VideoFrame.from_ndarray(np.array(frame.convert('RGB')), format='rgb24')
        for packet in self.stream.encode(av_frame):
            self.container.mux(packet)

    def value(self) -> pxt.Video:
        for packet in self.stream.encode():
            self.container.mux(packet)
        self.container.close()
        return str(self.out_file)


@pxt.udf(is_method=True)
def extract_audio(
    video_path: pxt.Video, stream_idx: int = 0, format: str = 'wav', codec: Optional[str] = None
) -> pxt.Audio:
    """
    Extract an audio stream from a video file, save it as a media file and return its path.

    Args:
        stream_idx: Index of the audio stream to extract.
        format: The target audio format. (`'wav'`, `'mp3'`, `'flac'`).
        codec: The codec to use for the audio stream. If not provided, a default codec will be used.
    """
    if format not in _format_defaults:
        raise ValueError(f'extract_audio(): unsupported audio format: {format}')
    default_codec, ext = _format_defaults[format]

    with av.open(video_path) as container:
        if len(container.streams.audio) <= stream_idx:
            return None
        audio_stream = container.streams.audio[stream_idx]
        # create this in our tmp directory, so it'll get cleaned up if it's being generated as part of a query
        output_filename = str(env.Env.get().tmp_dir / f'{uuid.uuid4()}.{ext}')

        with av.open(output_filename, 'w', format=format) as output_container:
            output_stream = output_container.add_stream(codec or default_codec)
            assert isinstance(output_stream, av.audio.stream.AudioStream)
            for packet in container.demux(audio_stream):
                for frame in packet.decode():
                    output_container.mux(output_stream.encode(frame))  # type: ignore[arg-type]

        return output_filename


@pxt.udf(is_method=True)
def get_metadata(video: pxt.Video) -> dict:
    """
    Gets various metadata associated with a video file and returns it as a dictionary.
    """
    return _get_metadata(video)


def _get_metadata(path: str) -> dict:
    with av.open(path) as container:
        assert isinstance(container, av.container.InputContainer)
        streams_info = [__get_stream_metadata(stream) for stream in container.streams]
        result = {
            'bit_exact': getattr(container, 'bit_exact', False),
            'bit_rate': container.bit_rate,
            'size': container.size,
            'metadata': container.metadata,
            'streams': streams_info,
        }
    return result


def __get_stream_metadata(stream: av.stream.Stream) -> dict:
    if stream.type not in ('audio', 'video'):
        return {'type': stream.type}  # Currently unsupported

    codec_context = stream.codec_context
    codec_context_md: dict[str, Any] = {
        'name': codec_context.name,
        'codec_tag': codec_context.codec_tag.encode('unicode-escape').decode('utf-8'),
        'profile': codec_context.profile,
    }
    metadata = {
        'type': stream.type,
        'duration': stream.duration,
        'time_base': float(stream.time_base) if stream.time_base is not None else None,
        'duration_seconds': float(stream.duration * stream.time_base)
        if stream.duration is not None and stream.time_base is not None
        else None,
        'frames': stream.frames,
        'metadata': stream.metadata,
        'codec_context': codec_context_md,
    }

    if stream.type == 'audio':
        # Additional metadata for audio
        channels = getattr(stream.codec_context, 'channels', None)
        codec_context_md['channels'] = int(channels) if channels is not None else None
    else:
        assert stream.type == 'video'
        assert isinstance(stream, av.video.stream.VideoStream)
        # Additional metadata for video
        codec_context_md['pix_fmt'] = getattr(stream.codec_context, 'pix_fmt', None)
        metadata.update(
            **{
                'width': stream.width,
                'height': stream.height,
                'frames': stream.frames,
                'average_rate': float(stream.average_rate) if stream.average_rate is not None else None,
                'base_rate': float(stream.base_rate) if stream.base_rate is not None else None,
                'guessed_rate': float(stream.guessed_rate) if stream.guessed_rate is not None else None,
            }
        )

    return metadata


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import abc
import itertools
import logging
import time
from dataclasses import dataclass
from typing import Any, Optional
from uuid import UUID

import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import Column, Table
from pixeltable.catalog import TableVersion, TableVersionHandle

_logger = logging.getLogger('pixeltable')


class ExternalStore(abc.ABC):
    """
    Abstract base class that represents an external data store that is linked to a Pixeltable
    table. Subclasses of `ExternalStore` provide functionality for synchronizing between Pixeltable
    and stateful external stores.
    """

    def __init__(self, name: str) -> None:
        self.__name = name

    @property
    def name(self) -> str:
        return self.__name

    @abc.abstractmethod
    def link(self, tbl_version: TableVersion) -> None:
        """
        Called by `TableVersion.link()` to implement store-specific logic.
        """

    @abc.abstractmethod
    def unlink(self, tbl_version: TableVersion) -> None:
        """
        Called by `TableVersion.unlink()` to implement store-specific logic.
        """

    @abc.abstractmethod
    def get_local_columns(self) -> list[Column]:
        """
        Gets a list of all local (Pixeltable) columns that are associated with this external store.
        """

    @abc.abstractmethod
    def sync(self, t: Table, export_data: bool, import_data: bool) -> SyncStatus:
        """
        Called by `Table.sync()` to implement store-specific synchronization logic.
        """

    @abc.abstractmethod
    def as_dict(self) -> dict[str, Any]: ...

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, md: dict[str, Any]) -> ExternalStore: ...


class Project(ExternalStore, abc.ABC):
    """
    An `ExternalStore` that represents a labeling project. Extends `ExternalStore` with a few
    additional capabilities specific to such projects.
    """

    stored_proxies: dict[Column, Column]

    def __init__(self, name: str, col_mapping: dict[Column, str], stored_proxies: Optional[dict[Column, Column]]):
        super().__init__(name)
        self._col_mapping = col_mapping

        # A mapping from original columns to proxy columns. A proxy column is an identical copy of a column that is
        # guaranteed to be stored; the Project will dynamically create and tear down proxy columns as needed. There
        # are two reasons this might happen:
        # (i) to force computed media data to be persisted; or
        # (ii) to force media data to be materialized in a particular location.
        # For each entry (k, v) in the dict, `v` is the stored proxy column for `k`. The proxy column `v` will
        # necessarily be a column of the table to which this project is linked, but `k` need not be; it might be a
        # column of a base table.
        # Note from aaron-siegel: This methodology is inefficient in the case where a table has many views with a high
        # proportion of overlapping rows, all proxying the same base column.
        if stored_proxies is None:
            self.stored_proxies: dict[Column, Column] = {}
        else:
            self.stored_proxies = stored_proxies

    def get_local_columns(self) -> list[Column]:
        return list(self.col_mapping.keys())

    def link(self, tbl_version: TableVersion) -> None:
        # All of the media columns being linked need to either be stored computed columns, or else have stored proxies.
        # This ensures that the media in those columns resides in the media store.
        # First determine which columns (if any) need stored proxies, but don't have one yet.
        stored_proxies_needed: list[Column] = []
        for col in self.col_mapping:
            if col.col_type.is_media_type() and not (col.is_stored and col.is_computed):
                # If this column is already proxied in some other Project, use the existing proxy to avoid
                # duplication. Otherwise, we'll create a new one.
                for store in tbl_version.external_stores.values():
                    if isinstance(store, Project) and col in store.stored_proxies:
                        self.stored_proxies[col] = store.stored_proxies[col]
                        break
                if col not in self.stored_proxies:
                    # We didn't find it in an existing Project
                    stored_proxies_needed.append(col)

        if len(stored_proxies_needed) > 0:
            _logger.info(f'Creating stored proxies for columns: {[col.name for col in stored_proxies_needed]}')
            # Create stored proxies for columns that need one. Increment the schema version
            # accordingly.
            tbl_version.version += 1
            preceding_schema_version = tbl_version.schema_version
            tbl_version.schema_version = tbl_version.version
            proxy_cols = [self.create_stored_proxy(tbl_version, col) for col in stored_proxies_needed]
            # Add the columns; this will also update table metadata.
            tbl_version._add_columns(proxy_cols, print_stats=False, on_error='ignore')
            # We don't need to retain `UpdateStatus` since the stored proxies are intended to be
            # invisible to the user.
            tbl_version._update_md(time.time(), preceding_schema_version=preceding_schema_version)

    def unlink(self, tbl_version: TableVersion) -> None:
        # Determine which stored proxies can be deleted. (A stored proxy can be deleted if it is not referenced by
        # any *other* external store for this table.)
        deletions_needed: set[Column] = set(self.stored_proxies.values())
        for name, store in tbl_version.external_stores.items():
            if isinstance(store, Project) and name != self.name:
                deletions_needed = deletions_needed.difference(set(store.stored_proxies.values()))
        if len(deletions_needed) > 0:
            _logger.info(f'Removing stored proxies for columns: {[col.name for col in deletions_needed]}')
            # Delete stored proxies that are no longer needed.
            tbl_version.version += 1
            preceding_schema_version = tbl_version.schema_version
            tbl_version.schema_version = tbl_version.version
            tbl_version._drop_columns(deletions_needed)
            self.stored_proxies.clear()
            tbl_version._update_md(time.time(), preceding_schema_version=preceding_schema_version)

    def create_stored_proxy(self, tbl_version: TableVersion, col: Column) -> Column:
        """
        Creates a proxy column for the specified column. The proxy column will be created in the specified
        `TableVersion`.
        """
        from pixeltable import exprs

        assert (
            col.col_type.is_media_type() and not (col.is_stored and col.is_computed) and col not in self.stored_proxies
        )
        proxy_col = Column(
            name=None,
            # Force images in the proxy column to be materialized inside the media store, in a normalized format.
            # TODO(aaron-siegel): This is a temporary solution and it will be replaced by a proper `destination`
            #   parameter for computed columns. Among other things, this solution does not work for video or audio.
            #   Once `destination` is implemented, it can be replaced with a simple `ColumnRef`.
            computed_with=exprs.ColumnRef(col).apply(lambda x: x, col_type=col.col_type),
            stored=True,
            col_id=tbl_version.next_col_id,
            sa_col_type=col.col_type.to_sa_type(),
            schema_version_add=tbl_version.schema_version,
        )
        proxy_col.tbl = TableVersionHandle(tbl_version.id, tbl_version.effective_version, tbl_version=tbl_version)
        tbl_version.next_col_id += 1
        self.stored_proxies[col] = proxy_col
        return proxy_col

    @property
    def col_mapping(self) -> dict[Column, str]:
        return self._col_mapping

    @abc.abstractmethod
    def get_export_columns(self) -> dict[str, ts.ColumnType]:
        """
        Returns the names and Pixeltable types that this `Project` expects to see in a data export. The keys
        of the `dict` are the names of data fields in the external store, not Pixeltable columns.

        Returns:
            A `dict` mapping names of external data fields to their expected Pixeltable types.
        """

    @abc.abstractmethod
    def get_import_columns(self) -> dict[str, ts.ColumnType]:
        """
        Returns the names and Pixeltable types that this `Project` provides in a data import.

        Returns:
            A `dict` mapping names of provided columns to their Pixeltable types.
        """

    @abc.abstractmethod
    def delete(self) -> None:
        """
        Deletes this `Project` and all associated (externally stored) data.
        """

    @classmethod
    def validate_columns(
        cls,
        table: Table,
        export_cols: dict[str, ts.ColumnType],
        import_cols: dict[str, ts.ColumnType],
        col_mapping: Optional[dict[str, str]],
    ) -> dict[Column, str]:
        """
        Verifies that the specified `col_mapping` is valid. In particular, checks that:
        (i) the keys of `col_mapping` are valid columns of the specified `Table`;
        (ii) the values of `col_mapping` are valid external columns (i.e., they appear in either `export_cols` or
            `import_cols`; and
        (iii) the Pixeltable types of the `col_mapping` keys are consistent with the expected types of the corresponding
            external (import or export) columns.
        If validation fails, an exception will be raised. If validation succeeds, a new mapping will be returned
        in which the Pixeltable column names are resolved to the corresponding `Column` objects.
        """
        from pixeltable import exprs

        is_user_specified_col_mapping = col_mapping is not None
        if col_mapping is None:
            col_mapping = {col: col for col in itertools.chain(export_cols.keys(), import_cols.keys())}

        resolved_col_mapping: dict[Column, str] = {}

        # Validate names
        t_cols = set(table._schema.keys())
        for t_col, ext_col in col_mapping.items():
            if t_col not in t_cols:
                if is_user_specified_col_mapping:
                    raise excs.Error(
                        f'Column name `{t_col}` appears as a key in `col_mapping`, but Table `{table._name}` '
                        'contains no such column.'
                    )
                else:
                    raise excs.Error(
                        f'Column `{t_col}` does not exist in Table `{table._name}`. Either add a column `{t_col}`, '
                        f'or specify a `col_mapping` to associate a different column with '
                        f'the external field `{ext_col}`.'
                    )
            if ext_col not in export_cols and ext_col not in import_cols:
                raise excs.Error(
                    f'Column name `{ext_col}` appears as a value in `col_mapping`, but the external store '
                    f'configuration has no column `{ext_col}`.'
                )
            col_ref = table[t_col]
            assert isinstance(col_ref, exprs.ColumnRef)
            resolved_col_mapping[col_ref.col] = ext_col
        # Validate column specs
        t_col_types = table._schema
        for t_col, ext_col in col_mapping.items():
            t_col_type = t_col_types[t_col]
            if ext_col in export_cols:
                # Validate that the table column can be assigned to the external column
                ext_col_type = export_cols[ext_col]
                if not ext_col_type.is_supertype_of(t_col_type, ignore_nullable=True):
                    raise excs.Error(
                        f'Column `{t_col}` cannot be exported to external column `{ext_col}` '
                        f'(incompatible types; expecting `{ext_col_type}`)'
                    )
            if ext_col in import_cols:
                # Validate that the external column can be assigned to the table column
                if table._tbl_version_path.get_column(t_col).is_computed:
                    raise excs.Error(
                        f'Column `{t_col}` is a computed column, which cannot be populated from an external column'
                    )
                ext_col_type = import_cols[ext_col]
                if not t_col_type.is_supertype_of(ext_col_type, ignore_nullable=True):
                    raise excs.Error(
                        f'Column `{t_col}` cannot be imported from external column `{ext_col}` '
                        f'(incompatible types; expecting `{ext_col_type}`)'
                    )
        return resolved_col_mapping

    @classmethod
    def _column_as_dict(cls, col: Column) -> dict[str, Any]:
        return {'tbl_id': str(col.tbl.id), 'col_id': col.id}

    @classmethod
    def _column_from_dict(cls, d: dict[str, Any]) -> Column:
        from pixeltable.catalog import Catalog

        tbl_id = UUID(d['tbl_id'])
        col_id = d['col_id']
        return Catalog.get().get_tbl_version(tbl_id, None).cols_by_id[col_id]


@dataclass(frozen=True)
class SyncStatus:
    external_rows_created: int = 0
    external_rows_deleted: int = 0
    external_rows_updated: int = 0
    pxt_rows_updated: int = 0
    num_excs: int = 0

    def combine(self, other: 'SyncStatus') -> 'SyncStatus':
        return SyncStatus(
            external_rows_created=self.external_rows_created + other.external_rows_created,
            external_rows_deleted=self.external_rows_deleted + other.external_rows_deleted,
            external_rows_updated=self.external_rows_updated + other.external_rows_updated,
            pxt_rows_updated=self.pxt_rows_updated + other.pxt_rows_updated,
            num_excs=self.num_excs + other.num_excs,
        )

    @classmethod
    def empty(cls) -> 'SyncStatus':
        return SyncStatus(0, 0, 0, 0, 0)


class MockProject(Project):
    """A project that cannot be synced, used mainly for testing."""

    def __init__(
        self,
        name: str,
        export_cols: dict[str, ts.ColumnType],
        import_cols: dict[str, ts.ColumnType],
        col_mapping: dict[Column, str],
        stored_proxies: Optional[dict[Column, Column]] = None,
    ):
        super().__init__(name, col_mapping, stored_proxies)
        self.export_cols = export_cols
        self.import_cols = import_cols
        self.__is_deleted = False

    @classmethod
    def create(
        cls,
        t: Table,
        name: str,
        export_cols: dict[str, ts.ColumnType],
        import_cols: dict[str, ts.ColumnType],
        col_mapping: Optional[dict[str, str]] = None,
    ) -> 'MockProject':
        col_mapping = cls.validate_columns(t, export_cols, import_cols, col_mapping)
        return cls(name, export_cols, import_cols, col_mapping)

    def get_export_columns(self) -> dict[str, ts.ColumnType]:
        return self.export_cols

    def get_import_columns(self) -> dict[str, ts.ColumnType]:
        return self.import_cols

    def sync(self, t: Table, export_data: bool, import_data: bool) -> SyncStatus:
        raise NotImplementedError()

    def delete(self) -> None:
        self.__is_deleted = True

    @property
    def is_deleted(self) -> bool:
        return self.__is_deleted

    def as_dict(self) -> dict[str, Any]:
        return {
            'name': self.name,
            'export_cols': {k: v.as_dict() for k, v in self.export_cols.items()},
            'import_cols': {k: v.as_dict() for k, v in self.import_cols.items()},
            'col_mapping': [[self._column_as_dict(k), v] for k, v in self.col_mapping.items()],
            'stored_proxies': [
                [self._column_as_dict(k), self._column_as_dict(v)] for k, v in self.stored_proxies.items()
            ],
        }

    @classmethod
    def from_dict(cls, md: dict[str, Any]) -> MockProject:
        return cls(
            md['name'],
            {k: ts.ColumnType.from_dict(v) for k, v in md['export_cols'].items()},
            {k: ts.ColumnType.from_dict(v) for k, v in md['import_cols'].items()},
            {cls._column_from_dict(entry[0]): entry[1] for entry in md['col_mapping']},
            {cls._column_from_dict(entry[0]): cls._column_from_dict(entry[1]) for entry in md['stored_proxies']},
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MockProject):
            return False
        return self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return f'MockProject `{self.name}`'



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Anthropic API. In order to use them, you must
first `pip install anthropic` and configure your Anthropic credentials, as described in
the [Working with Anthropic](https://pixeltable.readme.io/docs/working-with-anthropic) tutorial.
"""

import datetime
import json
import logging
from typing import TYPE_CHECKING, Any, Iterable, Optional, TypeVar, Union, cast

import httpx

import pixeltable as pxt
from pixeltable import env, exprs
from pixeltable.func import Tools
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import anthropic

_logger = logging.getLogger('pixeltable')


@env.register_client('anthropic')
def _(api_key: str) -> 'anthropic.AsyncAnthropic':
    import anthropic

    return anthropic.AsyncAnthropic(
        api_key=api_key,
        # recommended to increase limits for async client to avoid connection errors
        http_client=httpx.AsyncClient(limits=httpx.Limits(max_keepalive_connections=100, max_connections=500)),
    )


def _anthropic_client() -> 'anthropic.AsyncAnthropic':
    return env.Env.get().get_client('anthropic')


class AnthropicRateLimitsInfo(env.RateLimitsInfo):
    def __init__(self) -> None:
        super().__init__(self._get_request_resources)

    def _get_request_resources(self, messages: dict, max_tokens: int) -> dict[str, int]:
        input_len = 0
        for message in messages:
            if 'role' in message:
                input_len += len(message['role'])
            if 'content' in message:
                input_len += len(message['content'])
        return {'requests': 1, 'input_tokens': int(input_len / 4), 'output_tokens': max_tokens}

    def get_retry_delay(self, exc: Exception) -> Optional[float]:
        import anthropic

        # deal with timeouts separately, they don't come with headers
        if isinstance(exc, anthropic.APITimeoutError):
            return 1.0

        if not isinstance(exc, anthropic.APIStatusError):
            return None
        _logger.debug(f'headers={exc.response.headers}')
        should_retry_str = exc.response.headers.get('x-should-retry', '')
        if should_retry_str.lower() != 'true':
            return None
        retry_after_str = exc.response.headers.get('retry-after', '1')
        return int(retry_after_str)


@pxt.udf
async def messages(
    messages: list[dict[str, str]],
    *,
    model: str,
    max_tokens: int = 1024,
    metadata: Optional[dict[str, Any]] = None,
    stop_sequences: Optional[list[str]] = None,
    system: Optional[str] = None,
    temperature: Optional[float] = None,
    tool_choice: Optional[dict] = None,
    tools: Optional[list[dict]] = None,
    top_k: Optional[int] = None,
    top_p: Optional[float] = None,
    timeout: Optional[float] = None,
) -> dict:
    """
    Create a Message.

    Equivalent to the Anthropic `messages` API endpoint.
    For additional details, see: <https://docs.anthropic.com/en/api/messages>

    Request throttling:
    Uses the rate limit-related headers returned by the API to throttle requests adaptively, based on available
    request and token capacity. No configuration is necessary.

    __Requirements:__

    - `pip install anthropic`

    Args:
        messages: Input messages.
        model: The model that will complete your prompt.

    For details on the other parameters, see: <https://docs.anthropic.com/en/api/messages>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `claude-3-haiku-20240307`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> msgs = [{'role': 'user', 'content': tbl.prompt}]
        ... tbl.add_computed_column(response= messages(msgs, model='claude-3-haiku-20240307'))
    """

    # it doesn't look like count_tokens() actually exists in the current version of the library

    if tools is not None:
        # Reformat `tools` into Anthropic format
        tools = [
            {
                'name': tool['name'],
                'description': tool['description'],
                'input_schema': {
                    'type': 'object',
                    'properties': tool['parameters']['properties'],
                    'required': tool['required'],
                },
            }
            for tool in tools
        ]

    tool_choice_: Optional[dict] = None
    if tool_choice is not None:
        if tool_choice['auto']:
            tool_choice_ = {'type': 'auto'}
        elif tool_choice['required']:
            tool_choice_ = {'type': 'any'}
        else:
            assert tool_choice['tool'] is not None
            tool_choice_ = {'type': 'tool', 'name': tool_choice['tool']}
        if not tool_choice['parallel_tool_calls']:
            tool_choice_['disable_parallel_tool_use'] = True

    # make sure the pool info exists prior to making the request
    resource_pool_id = f'rate-limits:anthropic:{model}'
    rate_limits_info = env.Env.get().get_resource_pool_info(resource_pool_id, AnthropicRateLimitsInfo)
    assert isinstance(rate_limits_info, env.RateLimitsInfo)

    # TODO: timeouts should be set system-wide and be user-configurable
    from anthropic.types import MessageParam

    # cast(Any, ...): avoid mypy errors
    result = await _anthropic_client().messages.with_raw_response.create(
        messages=cast(Iterable[MessageParam], messages),
        model=model,
        max_tokens=max_tokens,
        metadata=_opt(cast(Any, metadata)),
        stop_sequences=_opt(stop_sequences),
        system=_opt(system),
        temperature=_opt(cast(Any, temperature)),
        tools=_opt(cast(Any, tools)),
        tool_choice=_opt(cast(Any, tool_choice_)),
        top_k=_opt(top_k),
        top_p=_opt(top_p),
        timeout=_opt(timeout),
    )

    requests_limit_str = result.headers.get('anthropic-ratelimit-requests-limit')
    requests_limit = int(requests_limit_str) if requests_limit_str is not None else None
    requests_remaining_str = result.headers.get('anthropic-ratelimit-requests-remaining')
    requests_remaining = int(requests_remaining_str) if requests_remaining_str is not None else None
    requests_reset_str = result.headers.get('anthropic-ratelimit-requests-reset')
    requests_reset = datetime.datetime.fromisoformat(requests_reset_str.replace('Z', '+00:00'))
    input_tokens_limit_str = result.headers.get('anthropic-ratelimit-input-tokens-limit')
    input_tokens_limit = int(input_tokens_limit_str) if input_tokens_limit_str is not None else None
    input_tokens_remaining_str = result.headers.get('anthropic-ratelimit-input-tokens-remaining')
    input_tokens_remaining = int(input_tokens_remaining_str) if input_tokens_remaining_str is not None else None
    input_tokens_reset_str = result.headers.get('anthropic-ratelimit-input-tokens-reset')
    input_tokens_reset = datetime.datetime.fromisoformat(input_tokens_reset_str.replace('Z', '+00:00'))
    output_tokens_limit_str = result.headers.get('anthropic-ratelimit-output-tokens-limit')
    output_tokens_limit = int(output_tokens_limit_str) if output_tokens_limit_str is not None else None
    output_tokens_remaining_str = result.headers.get('anthropic-ratelimit-output-tokens-remaining')
    output_tokens_remaining = int(output_tokens_remaining_str) if output_tokens_remaining_str is not None else None
    output_tokens_reset_str = result.headers.get('anthropic-ratelimit-output-tokens-reset')
    output_tokens_reset = datetime.datetime.fromisoformat(output_tokens_reset_str.replace('Z', '+00:00'))
    retry_after_str = result.headers.get('retry-after')
    if retry_after_str is not None:
        _logger.debug(f'retry-after: {retry_after_str}')

    rate_limits_info.record(
        requests=(requests_limit, requests_remaining, requests_reset),
        input_tokens=(input_tokens_limit, input_tokens_remaining, input_tokens_reset),
        output_tokens=(output_tokens_limit, output_tokens_remaining, output_tokens_reset),
    )

    result_dict = json.loads(result.text)
    return result_dict


@messages.resource_pool
def _(model: str) -> str:
    return f'rate-limits:anthropic:{model}'


def invoke_tools(tools: Tools, response: exprs.Expr) -> exprs.InlineDict:
    """Converts an Anthropic response dict to Pixeltable tool invocation format and calls `tools._invoke()`."""
    return tools._invoke(_anthropic_response_to_pxt_tool_calls(response))


@pxt.udf
def _anthropic_response_to_pxt_tool_calls(response: dict) -> Optional[dict]:
    anthropic_tool_calls = [r for r in response['content'] if r['type'] == 'tool_use']
    if len(anthropic_tool_calls) == 0:
        return None
    pxt_tool_calls: dict[str, list[dict[str, Any]]] = {}
    for tool_call in anthropic_tool_calls:
        tool_name = tool_call['name']
        if tool_name not in pxt_tool_calls:
            pxt_tool_calls[tool_name] = []
        pxt_tool_calls[tool_name].append({'args': tool_call['input']})
    return pxt_tool_calls


_T = TypeVar('_T')


def _opt(arg: _T) -> Union[_T, 'anthropic.NotGiven']:
    import anthropic

    return arg if arg is not None else anthropic.NOT_GIVEN


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



# ruff: noqa: F401

from .catalog import Catalog
from .column import Column
from .dir import Dir
from .globals import IfExistsParam, IfNotExistsParam, MediaValidation, UpdateStatus, is_valid_identifier, is_valid_path
from .insertable_table import InsertableTable
from .named_function import NamedFunction
from .path import Path
from .schema_object import SchemaObject
from .table import Table
from .table_version import TableVersion
from .table_version_handle import TableVersionHandle
from .table_version_path import TableVersionPath
from .view import View



import dataclasses
import importlib
import logging
import os
import pkgutil
from typing import Callable

import sqlalchemy as sql
from sqlalchemy import orm

from pixeltable.utils.console_output import ConsoleLogger

from .schema import SystemInfo, SystemInfoMd

_console_logger = ConsoleLogger(logging.getLogger('pixeltable'))


# current version of the metadata; this is incremented whenever the metadata schema changes
VERSION = 31


def create_system_info(engine: sql.engine.Engine) -> None:
    """Create the system metadata record"""
    system_md = SystemInfoMd(schema_version=VERSION)
    record = SystemInfo(md=dataclasses.asdict(system_md))
    with orm.Session(engine, future=True) as session:
        # Write system metadata only once for idempotency
        if session.query(SystemInfo).count() == 0:
            session.add(record)
            session.flush()
            session.commit()


# conversion functions for upgrading the metadata schema from one version to the following
# key: old schema version
converter_cbs: dict[int, Callable[[sql.engine.Engine], None]] = {}


def register_converter(version: int) -> Callable[[Callable[[sql.engine.Engine], None]], None]:
    def decorator(fn: Callable[[sql.engine.Engine], None]) -> None:
        assert version not in converter_cbs
        converter_cbs[version] = fn

    return decorator


# load all converter modules
for _, modname, _ in pkgutil.iter_modules([os.path.dirname(__file__) + '/converters']):
    importlib.import_module('pixeltable.metadata.converters.' + modname)


def upgrade_md(engine: sql.engine.Engine) -> None:
    """Upgrade the metadata schema to the current version"""
    with orm.Session(engine) as session:
        system_info = session.query(SystemInfo).one().md
        md_version = system_info['schema_version']
        assert isinstance(md_version, int)
        if md_version == VERSION:
            return
        while md_version < VERSION:
            if md_version not in converter_cbs:
                raise RuntimeError(f'No metadata converter for version {md_version}')
            # We can't use the console logger in Env, because Env might not have been initialized yet.
            _console_logger.info(f'Converting metadata from version {md_version} to {md_version + 1}')
            converter_cbs[md_version](engine)
            md_version += 1
        # update system info
        conn = session.connection()
        system_info_md = SystemInfoMd(schema_version=VERSION)
        conn.execute(SystemInfo.__table__.update().values(md=dataclasses.asdict(system_info_md)))
        session.commit()
